# CasFinder.pl
#
# Perl script for finding candidate specific Cas9 targets in sequence
# provided by a user and for evaluating these for potential off-targets
# in a specified genome.  CasFinder contains logic for finding candidate
# sites in user-provided input sequence, and then calls the CasValue
# program to evaluate these for off-targets in a specified genome.
# After CasValue returns, it filters out candidate targets that did not
# meet CasValue's evaluation and prints out the results.  Note that while
# these comments refer to the "CasValue" program, CasFinder itself uses
# the variable $CASVALUE_PROGRAM is used to refer to the version of the
# program that CasFinder that it calls.  As of this writing, this program
# is CasValue_v2.pl. This variable is defined in a configuration file
# and can be overridden via a program option.
#
# A flow chart that presents the high-level logic of CasFinder is 
# available as Supporting Info, Figure S1 of the article describing
# the CasFinder system (see citation below).  Within the code below
# the notation "FLOWCHART NOTE" signifies a call or block of code that
# is represented in the flow chart.  Note that perl variable names in 
# CAPITAL LETTERS correspond to the data items in the box labeled 
# "CasFinder internal memory" in the flow chart.
#
# CasFinder and the CasValue program both use a pair of configuration
# files, one of which defines genome and system files and resources,
# and the other of which defines the characteristics of the Cas9s that
# will be recognized by the programs.  However, neither program accesses
# these files directly but rather through a pair of perl packages that
# encapsulate access to them.  These packages are CasFinderConfigFile.pm
# and CasFinderConfigCas9File.pm.  CasFinder gets information on the target
# sequence lengths and PAMs of defined Cas9s from this latter package.
#
# Detailed instructions on how to install and use CasFinder and
# CasValue, as well as other program documentation, can be found on
# http://arep.med.harvard.edu/CasFinder.  The following summarizes
# CasFinder's program logic:
#
# 1. Validate input program options, read configuration files, and
# consolidate this information.
#
# 2. Process the input sequence file. The CasFinder input file is in a
# modified FASTA format in which the FASTA headers are read for special
# keywords that control how Cas9 targets found in the sequence will be
# evaluated and reported.  The keywords and their usage can be found in
# the program documentation at the web site above.  In processing the
# input file, CasFinder first validates and saves the information in
# the headers in internal tables, and saves any sequences provided in
# the input file.  However, CasFinder allows some FASTA records to be
# supplied as headers alone without sequence: These contain the *EXTRACT*
# keyword along with other keywords that specify a location ranges in the
# genome for which CasValue will evaluate Cas9 targets for off-targets.
# After the contents of the input file are analyzed, CasFinder then
# reads through the chromosome files for this genome and itself extracts
# the sequences for which *EXTRACT* was specified.  All *EXTRACT*s are
# batched by chromosome, and chromosome files are read in one at a time,
# in order to reduce memory requirements.
#
# 3. CasFinder then searches these sequences for candidate Cas9 targets
# by looking for substrings ending in PAM motifs for the Cas9 it is
# processing. By default, CasFinder avoids any candidate targets that
# contain repeatmasked sequences indicated as lower-case base codes,
# but this can be overridden by keywords in the input sequence FASTA
# header or by use of the -R program option.  (When options dictate
# that repeatmasked sequence is to be analyzed for candidate targets,
# the input sequence is simply all converted to upper case.)
#
# 4. CasFinder then invokes the CasValue program to evaluate the candidate
# targets it has found.  It generates a CasFinder input file that comprises
# a FASTA file that contains only the targeting sequences of the candidate
# targets it has found (i.e., the PAM sequences are *not* passed to the
# CasValue program), and with keywords in the FASTA headers recognized
# by the CasValue program.  It then formats a string of CasValue program
# options and invokes $CASVALUE_PROGRAM with this string. The way in which
# CasValue passes options to the CasValue program is described in detail
# in the program documentation at the web site above.  The CasValue input
# file that CasFinder passes to CasValue, and the output files generated
# by CasValue, are all temporary files that will be deleted after CasFinder
# is through with them.
#
# 5. When the CasValue program completes, CasFinder reads through the
# CasValue output, and then deletes the temporary files set up for the
# CasValue program, with the exception of the 'info' file generated
# by CasValue, a copy of which is saved as a CasFinder output file.
# If the -D program option is used, a copy of this file is also saved.
# (Note that if the CasValue program terminates on a program error, there
# is a possibility that CasFinder will not be able to copy the CasValue
# 'info' or -D files.)  CasFinder then generates its own output files.
# Candidate sites that passed CasValues off-target specificity check will
# be reported in the output 'data' file, where reporting takes into account
# the instructions coded in the input sequence FASTA header keywords.
# Data file records will also include information passed back by the
# CasValue program such as threshold_rejection_score that can be used to
# rank reported sites by their the levels of specificity.
#
# USAGE: Help messages summarizing CasFinder.pl program options may be
# viewed by executing
#
#    perl CasFinder.pl -h
#
# from a linux command line.  Full details are provided in the program
# documentation at the web site above.
#
# CUSTOMIZATION: Full details are provided in the program documentation
# at the web site above.
#
# CasFinder invokes the CasValue program which invokes the bowtie sequence
# mapping program.  For information on bowtie, see
#
#  Langmead B, Trapnell C, Pop M, Salzberg SL.,
#     Ultrafast and memory-efficient alignment of short DNA sequences to
#     the human genome. Genome Biol 10:R25.
#
# AVAILABILITY AND CITATION
# 
# The CasFinder system of programs comprises CasFinder.pl, CasValue_v2.pl,
# the two perl packages CasFinderConfigFile.pm and CasFinderConfigCas9File.pm, 
# and the two configuration files CASFINDER_CONFIG.txt and 
# CASFINDER_CONFIG_CAS9.txt.
# 
# The CasFinder system of programs is being made available under the MIT 
# Licence (MIT).   
#
# Please cite the article: 
#
#   Aach J., Mali P., Church G.M..
#     CasFinder: Flexible algorithm for identifying specific Cas9 targets 
#     in genomes.  bioRxiv (doi)   
#
# when using, copying, or employing code from this system.
#
# COPYRIGHT NOTICE
# 
# The MIT License (MIT)
#
# Copyright (c) 2014 John Aach and the President and Fellows of Harvard University
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# PROGRAM CHANGE LOG:
# Created: 09/09/2013-05/12/2014 by John Aach
#

  use File::Spec;
  use Cwd 'abs_path';
  use File::Copy;

  @INC=(".",@INC);

  use CasFinderConfigFile;
  use CasFinderConfigCas9File;

  ($vol,$dir,$PROGRAM)=File::Spec->splitpath(abs_path($0));
  $PGMPREF=$PROGRAM;
  $PGMPREF=~s/\.pl$//;
  
  # FLOWCHART NOTE:
  # The following call performs the logic in the box "Analyze parameters & retrieve config file info"
  AnalyzeParameters(@ARGV);
  
  # FLOWCHART NOTE:
  # The following call performs the logic in the box "Read input sequence; analyze FASTA records"
  ProcessInputFile(); # generates @START, @END, @REFPOINT, @REFORIENT, @OTHER_ANNOTATIONS, @SKIP, @EXTRACT, @SEARCHREPEAT, @SEQ
  
  $CasValue_infile="$TEMPPREF.casfinder.casvalue.tgt.fa";
  $CasValue_outpref="$TEMPPREF.casfinder.casvalue.tgt";
  # FLOWCHART NOTE:
  # The following call performs the logic in the box "Search input sequence for candidate Cas9 targets"
  FindCandidateTargetSequences($CasValue_infile); # generates %TGT_CANDIDATES
  
  # FLOWCHART NOTE:
  # The following call performs the logic in the box "Invoke CasValue to check Cas9 target specificity"
  $normal_completion=RunCasValueProgram($CasValue_infile,$CasValue_outpref);
  $CasValue_output_datafile="$CasValue_outpref.data.txt";
  if($normal_completion==0){
     Msg("CasFinder terminating due to $CASVALUE_PGMNAME error.\n");
     if($DEBUG==0){
       unlink glob "$TEMPPREF.*";  
       unlink $CasValue_output_datafile;
     }
     close I;
     exit;
  }
  
  # FLOWCHART NOTE:
  # The following call performs the logic in the box "Report accepted Cas9 targets"
  ReportTargets($CasValue_output_datafile);

  Msg("\nProgram complete.");
  close I;
  if($DEBUG==0){
    unlink glob "$TEMPPREF.*";
  }
  

# FLOWCHART NOTE:
# The following subroutine performs the logic in the box "Read input file sequence; analyze FASTA keywords"
sub ProcessInputFile{
  my $rgn_no;
  my $id;
  my $rest;
  my $extract;
  my $chr;
  my $start;
  my $end;
  my $searchrepeat;
  my $refpoint;
  my $reforient;
  my $keyword;
  my $accept;
  my $fastalen=0;
  my $extractlen=0;
  my $i;
  my $numseq;
  my $skipno=0;
  my $extractno=0;
  my %chrom_extract_info=();
  my $arr;
  my $chrhdr;
  my $chrseq="";
  my $extract_start;
  my $extract_lnth;
  my $extract_ix;

  Msg("\nProcessing input file $INFILE\n");
  open(IN,$INFILE);
  $rgn_no=-1;
  while(<IN>){
    s/\s+$//;
    if(/^>(\S+)(.*)/){
       $id=$1;
       $rest=$2;
       $extract=0;
       ++$rgn_no;
       $ID[$rgn_no]=$id;
       if($rest=~/\*extract\*/i){  # IS THIS AN EXTRACTED SEQUENCE??
         $EXTRACT[$rgn_no]=$extract=1;
         $rest=~s/\s*\*extract\*\s*/ /ig;
       }
       else {
          $START[$rgn_no]=1;
          $END[$rgn_no]=0;  # will calculate below from sequence length
       } 
       if($extract==1 && $rest=~/range=([^:]+):(\d+)-(\d+)/i){
         $chr=$1;
         $start=$2;
         $end=$3;
         $CHR[$rgn_no]=$chr;
         $START[$rgn_no]=$start;
         $END[$rgn_no]=$end;
         $rest=~s/\s*range=[^:]+:\d+-\d+\s*/ /ig;
       }
       if($rest=~/searchrepeat=((Y|N))/i){
         $searchrepeat=uc($1);
         if($searchrepeat eq "N"){
           $SEARCHREPEAT[$rgn_no]=0;
         }
         else {
           $SEARCHREPEAT[$rgn_no]=1;
         }
         $rest=~s/\s*searchrepeat=(Y|N)\s*/ /ig;
       }
       if($rest=~/refpoint=(\d+)/i){
         $REFPOINT[$rgn_no]=$refpoint=$1;
         $rest=~s/\s*refpoint=\d+\s*/ /ig;
       }
       if($rest=~/reforient=(\+|-)/i){
         $REFORIENT[$rgn_no]=$reforient=$1;
         $keyword="\\s*reforient=".quotemeta($reforient)."\\s*";
         $rest=~s/$keyword/ /ig;
       }
       if($rest=~/accept=(\d+)/i){
         $ACCEPTNUM[$rgn_no]=$accept=$1;
         $rest=~s/\s*accept=\d+\s*/ /ig;
       }
       unless($rest=~/^\s+$/){
         $rest=~s/\t/ /g; # get rid of any tabs
         $rest=~s/^\s+//; # eliminate leading and trailing blanks
         $rest=~s/\s+$//;
         if($rest=~/\s/||$rest=~/"/){ # put quotes around this if there are blanks or internal quotes.
            $rest=~s/"/""/g;
            $rest="\"$rest\"";
         }
         $OTHER_ANNOTATION[$rgn_no]=$rest;
       }
       if($ACCEPTNUM[$rgn_no] eq ""){  # set default if not otherwise specified.
         if($EXTRACT[$rgn_no]==0){
           $ACCEPTNUM[$rgn_no]=0;
         }
         else{
           $ACCEPTNUM[$rgn_no]=1;
         }
       }
    }
    else {
       $SEQ[$rgn_no].=$_;
       if($extract==0){
         $END[$rgn_no]=length($SEQ[$rgn_no]);  # calculating end of non-extracted entries from length
       }
    }
  }
  close IN;

  $numseq=$rgn_no+1;
  Msg("$numseq sequences specified in input file. Performing initial validations\n");
  for($i=0;$i<$numseq;++$i){
    # set some defaults
    $SKIP[$i]=0; 
    if($REFPOINT[$i] eq""){ # default changed to START: 02/14/2014
       $REFPOINT[$i]=$START[$i];
    } 
    if($REFORIENT[$i] eq""){ $REFORIENT[$i]="+";}
    if($FORCE_SEARCHREPEAT==1){ # override individual specifications if -R parameter provided. 
      $SEARCHREPEAT[$i]=1; 
    }
    elsif($SEARCHREPEAT[$i]eq""){ # otherwise, default is to avoid searching
      $SEARCHREPEAT[$i]=0; 
    }   
    if($START[$i]>$END[$i] || $START[$i]==0 || $END[$i]==0){
       MsgContinue("Error in input sequence $ID[$i] (input file index $i): invalid sequence range start or end point.");
       $SKIP[$i]=1;
       $skipno++;
       next;
    }  
    if($EXTRACT[$i]==1 && $SEQ[$i] ne ""){
       MsgContinue("Error in input sequence $ID[$i] (input file index $i): sequence provided for *extract* input file entry.");
       $SKIP[$i]=1;
       $skipno++;
       next;
    }  
    if($EXTRACT[$i]==0 && $SEQ[$i] eq ""){
       MsgContinue("Error in input sequence $ID[$i] (input file index $i): no sequence provided for FASTA input file entry.");
       $SKIP[$i]=1;
       $skipno++;
       next;
    }  
    if($REFPOINT[$i]< $START[$i] || $REFPOINT[$i]>$END[$i]){
       MsgContinue("Error in input sequence $ID[$i] (input file index $i): refpoint inconsistent with sequence range or length.");
       $SKIP[$i]=1;
       $skipno++;
       next;
    }  

    # accumulate fastseq length to be analyzed
    if($EXTRACT[$i]==0){
       $fastalen+=length($SEQ[$i]);
    }
  }
  Msg("\n$skipno sequences will not be processed due to specification validity errors.\n");

  # FLOWCHART NOTE:
  # The logic that follows  performs the logic in the box "Extract sequences from chromosome files as requested"

  # prepare to process sequence extractions
  for($i=0;$i<$numseq;++$i){
    if($EXTRACT[$i]==1 && $SKIP[$i]==0){
       if($chrom_extract_info{$CHR[$i]} eq""){
         $chrom_extract_info{$CHR[$i]}=Array();
       }
       $arr=$chrom_extract_info{$CHR[$i]};
       push @$arr,(join"\t",($START[$i]-1,$END[$i]-$START[$i]+1,$i)); 
       $extractno++;
    }
  }

  if($extractno>0){
    Msg("\nExtracting $extractno sequences from the $GENOME reference sequence.\n");

    foreach $chr(@$CURR_GENOME_CHROMOSOME_ORDER){
      if($chrom_extract_info{$chr} eq ""){ next;} # nothing for this chromosome
      $arr=$chrom_extract_info{$chr};
      open(C,$$CURR_GENOME_CHROMOSOME_FILES{$chr});
      $chrhdr=<C>;
      $chrseq="";
      while(<C>){
         s/\s$+//;
         if(/^>/){
            Msg("\nUnexpected FASTA header in chromosome file $$CURR_GENOME_CHROMOSOME_FILES{$chr}:");
            MsgContinue("$_");
            MsgContinue("Program terminating.");
            exit;
         }
         $chrseq.=$_;
      }
      close C;
      # extract sequences
      for($i=0;$i<@$arr;++$i){
        ($extract_start,$extract_lnth,$extract_ix)=split/\t/,$$arr[$i];
        $SEQ[$extract_ix]=substr($chrseq,$extract_start,$extract_lnth);
        $extractlen+=length($SEQ[$extract_ix]);  
      }
      delete $chrom_extract_info{$chr};   # get rid of unused storage
    }
  }

  Msg("\nInput file processed: Total FASTA sequence length = $fastalen, total extracted sequence length = $extractlen\n"); 
  
}

# FLOWCHART NOTE:
# The following subroutine performs the logic in the box "Search input sequence for candidate Cas9 targets"
sub FindCandidateTargetSequences{
  my $CasValue_infile=$_[0];
  my $rgn_no;
  my $pam_re_query_string=$CURR_CAS9_PAM_REGEXP;  
  my $rcpam_re_query_string=CasFinderConfigCas9File->RCREGEXP($pam_re_query_string); 
  my $tgt;
  my $tgtpam;
  my $tgtpref;
  my $tgtname;
  my $tgtlnth=$CURR_CAS9_TARGETLEN;
  my $full_lnth=$tgtlnth+$CURR_CAS9_PAMLNTH;       
  my $pos;
  my $cas9tgt_5p;
  my $refoffset;
  my $i;
  my $seq;
  my $tstring="T"x$TFILTER;
  my $trejcount=0;
  my $nonstdbase_rejcount=0;
  my $tot_candidates=0;
  my $dbgix=0;
  my $dbglimit=50;

  Msg("Beginning input search for candidate Cas9 targets in sequence data.\n");

  %TGT_CANDIDATES=();

  open(CANDTARGET,">$CasValue_infile");  # initial candidate seed file

  if($DEBUG==1){
    print DEBUG "$dbglimit sequences from each strand:\n";
    print DEBUG "PAM regexp = $pam_re_query_string, PAM rc-regexp = $rcpam_re_query_string\n";
  }

  # The logic for searching for Cas9 candidate targets is done twice, once on the + strand
  # and then again on the - strand.  Different handling of the two strands is needed because
  # different search patterns are used (+ sense vs - sense PAM patterns), and because 
  # positions and offsets need to be computed differentl.  When a candidate is found, 
  # its targeting sequence is written as a FASTA record to a file that will be used as
  # an input file to CasFinder.  A temporary name is assigned to the target site that 
  # is based on the index of the input sequence in the CasFinder input file in which 
  # it was found, the position of the target in the input sequence, and the suffix "r"
  # if the target was found on the reverse strand.

  for($rgn_no=0;$rgn_no<@SEQ;++$rgn_no){
    if($SKIP[$rgn_no]==1){ next;}
    $seq=$SEQ[$rgn_no];
    if($SEARCHREPEAT[$rgn_no]==1){
       $seq=uc($seq);
    }
    $tgtpref="R$rgn_no";
    while($seq=~m/$pam_re_query_string/g){  
       if(pos($seq)<$full_lnth){next;}  
       $tgtpam=substr($seq,pos($seq)-$full_lnth,$full_lnth);  
       $pos=(pos $seq)-length($tgtpam)+1; # position of 5' base of target in region
       pos($seq)=pos($seq)-($CURR_CAS9_PAMLNTH-1);  
       if($tgtpam=~/[acgt]/){ next;}  # avoid searching repetitive region 
       if($tgtpam=~/[^ACGT]/){
         ++$nonstdbase_rejcount;
         next;
       }          
       $cas9tgt_5p=$START[$rgn_no]+$pos-1;
       $tgt=substr($tgtpam,0,$CURR_CAS9_TARGETLEN);
       if($tgt=~/$tstring/){
          ++$trejcount;
          if($DEBUG==1&&$dbgix>=0&&$dbgix<$dbglimit){
            print DEBUG "Forward strand T-string rejection: $tgtpam, $pos, $cas9tgt_5p\n";
          }
          next;
       }
       $tgtname="$tgtpref-$cas9tgt_5p";
       if($REFORIENT[$rgn_no] eq "+" || $REFORIENT[$rgn_no] eq ""){
         $refoffset=$cas9tgt_5p-$REFPOINT[$rgn_no];
       }
       else {
         $refoffset=-$cas9tgt_5p+$REFPOINT[$rgn_no];
       }
       $TGT_CANDIDATES{$tgtname}="$refoffset\t$tgtpam";
       if($DEBUG==1&&$dbgix<$dbglimit){
          ++$dbgix;
          print DEBUG "Forward strand $dbgix: $tgtname: $TGT_CANDIDATES{$tgtname}\t".
                "(start,end,refpoint,reforient,other,acceptnum)=".
                "($START[$rgn_no],$END[$rgn_no],$REFPOINT[$rgn_no],$REFORIENT[$rgn_no],$OTHER_ANNOTATION[$rgn_no],$ACCEPTNUM[$rgn_no])\n";
       }
       print CANDTARGET ">$tgtname accept=$ACCEPTNUM[$rgn_no]\n$tgt\n";
       $tot_candidates++;
    }
    if($DEBUG==1){
      $dbgix=0;
    }
    while($seq=~m/$rcpam_re_query_string/g){  
       $tgtpam=RC(substr($seq,pos($seq)-$CURR_CAS9_PAMLNTH,$full_lnth));   
       if(length($tgtpam)!=$full_lnth){ next; } 
       $pos=(pos $seq)+$tgtlnth; # position of 5' base of target in region 
       pos($seq)=pos($seq)-($CURR_CAS9_PAMLNTH-1);  
       if($tgtpam=~/[acgt]/){ next;}  # avoid searching repetitive region 
       if($tgtpam=~/[^ACGT]/){
         ++$nonstdbase_rejcount;
         next;
       }          
       $cas9tgt_5p=$START[$rgn_no]+$pos-1;
       $tgt=substr($tgtpam,0,$CURR_CAS9_TARGETLEN);
       if($tgt=~/$tstring/){
          ++$trejcount;
          if($DEBUG==1&&$dbgix>=0&&$dbgix<$dbglimit){
            print DEBUG "Reverse strand T-string rejection: $tgtpam, $pos, $cas9tgt_5p\n";
          }
          next;
       }
       $tgtname="$tgtpref-${cas9tgt_5p}r";
       if($REFORIENT[$rgn_no] eq "+" || $REFORIENT[$rgn_no] eq ""){
         $refoffset=$cas9tgt_5p-$REFPOINT[$rgn_no];
       }
       else {
         $refoffset=-$cas9tgt_5p+$REFPOINT[$rgn_no];
       }
       $TGT_CANDIDATES{$tgtname}="$refoffset\t$tgtpam";
       if($DEBUG==1&&$dbgix>=0&&$dbgix<$dbglimit){
          ++$dbgix;
          print DEBUG "Reverse strand $dbgix: $tgtname: $TGT_CANDIDATES{$tgtname}\t".
                "(start,end,refpoint,reforient,other,acceptnum)=".
                "($START[$rgn_no],$END[$rgn_no],$REFPOINT[$rgn_no],$REFORIENT[$rgn_no],$OTHER_ANNOTATION[$rgn_no],$ACCEPTNUM[$rgn_no])\n";
       }
       print CANDTARGET ">$tgtname accept=$ACCEPTNUM[$rgn_no]\n$tgt\n";
       $tot_candidates++;
    }
  }        
  close CANDTARGET;

  Msg("\n$tot_candidates candidate Cas9 targets identified from sequence data.");
  MsgContinue("$trejcount additional candidates were rejected due to T-filter.");
  MsgContinue("$nonstdbase_rejcount additional candidates were rejected due to non-standard base codes.\n");

}

# FLOWCHART NOTE:
# The following subroutine performs the logic in the box "Invoke CasValue to check Cas9 target specificity"
sub RunCasValueProgram{
  my $CasValue_infile=$_[0];
  my $CasValue_outpref=$_[1];
  my $casvalue_command;
  my $msg;
  my $complete;
  my $CasValue_info_file;
  my $CasValue_info_file_copy;
  my $CasValue_site_detail_file;  
  my $CasValue_site_detail_file_copy;  
  my $casvalue_k_spec="";  
  my $casvalue_x_spec="";  
  my $casvalue_D_spec="";  
  my $casvalue_A_spec="";  

  # FLOWCHART NOTE:
  # The logic that follows performs the logic in the box "Casll CasValue passing through program options"

  if($CASVALUE_K ne ""){ 
    $casvalue_k_spec="-k $CASVALUE_K";
  }
  if($CASVALUE_X ne ""){ 
    $casvalue_x_spec="-x $CASVALUE_X";
  }
  if($CASVALUE_A ne ""){ 
    $casvalue_A_spec="-A $CASVALUE_A";
  }
  if($CASVALUE_D==1){ 
    $casvalue_D_spec="-D";
  }

  $casvalue_command="perl $CASVALUE_PROGRAM -g $GENOME -c $CAS9 $casvalue_k_spec $casvalue_x_spec $casvalue_A_spec $casvalue_D_spec ".  
        "-CONFIG $CONFIG_FILE -CONFIG_CAS9 $CONFIG_CAS9_FILE " . 
        "-i $CasValue_infile -o $CasValue_outpref $CASVALUE_PARAMETER_STRING"; 
  if($DEBUG==1){
    print DEBUG "$CASVALUE_PGMNAME command: $casvalue_command\n";
  }
  Msg("Initiating program $CASVALUE_PGMNAME to evaluate candidate target specificity.\n");
  system($casvalue_command);
  # save copy of CasValue program info file and check for Program complete message
  $CasValue_info_file="$CasValue_outpref.info.txt";
  open(CVI,$CasValue_info_file);
  $complete=0;
  while(<CVI>){
    if(/^\d+:\d+:\d+:\s+Program complete\.?\s*$/i){
       $complete=1;
    }
  }
  close CVI;
  if($complete==1){
     $msg="and terminated normally.";
  }
  else {
    $msg="but did not terminate normally.";
  }
  Msg("\n$CASVALUE_PGMNAME has completed execution $msg");
  # Only report that the CasValue_v2's info file has been copied if the program has actually completed
  if(-e $CasValue_info_file){
    $CasValue_info_file_copy="$OUTPREF.$CASVALUE_PGMNAME.info.txt";
    copy($CasValue_info_file,$CasValue_info_file_copy);  # File::Copy
    MsgContinue("A copy of the $CASVALUE_PGMNAME info file has been saved as $CasValue_info_file_copy.");
  }
  else {
    MsgContinue("No copy of the $CASVALUE_PGMNAME info file was saved because this file could not be detected.");
  }
  # Now save a copy of the the CasValue_v2 site_detail file if one was created
  $CasValue_site_detail_file="$CasValue_outpref.site_detail.txt";
  if(-e $CasValue_site_detail_file){
    $CasValue_site_detail_file_copy="$OUTPREF.$CASVALUE_PGMNAME.site_detail.txt";
    my $copy_successful=CopySiteDetailFile($CasValue_site_detail_file,$CasValue_site_detail_file_copy);
    if($copy_successful==1){
       MsgContinue("A copy of the $CASVALUE_PGMNAME site_detail file has been saved as $CasValue_site_detail_file_copy.");   
       MsgContinue("Candidate target information in site_detail file was successfully resolved to input sequence names.");   
    }
    else {
       copy($CasValue_site_detail_file,$CasValue_site_detail_file_copy);  # File::Copy
       MsgContinue("A copy of the $CASVALUE_PGMNAME site_detail file has been saved as $CasValue_site_detail_file_copy.");   
       MsgContinue("The site_detail copy contains temporary candidate target names due to an error resolving these to input sequence names.");   
    }
  }
  elsif($CASVALUE_D==1) {
    MsgContinue("No copy of the $CASVALUE_PGMNAME site_detail file was saved because this file could not be detected.");
  }
   
  return $complete;
}

# This subroutine creates a copy of the CasValue_v2 site_detail file (if present) while translating temporary 
# target names to CasFinder input sequence names + sequence locations.
sub CopySiteDetailFile{
  my $CasValue_site_detail_file=$_[0];
  my $CasValue_site_detail_file_copy=$_[1];
  my $hdr;
  my @cols;
  my $ok;
  my $rgn_no;
  my $seqrange;
  my $target_range;
  my $target_strand;
  my $target_refpoint_5p_offset;
  my $target_sequence;
  my $targetname;
  my $copy_successful=1;
  open(SD,$CasValue_site_detail_file);
  open(SDC,">$CasValue_site_detail_file_copy");
  $hdr=<SD>;
  @cols=split/\t/,$hdr;
  # replace first two columns (target_index and target_name with 6 columns, $rgn_no+1, $ID[$rgn_no], $seqrange,$target_range,$target_strand, $target_sequence 
  print SDC (join"\t",("input_sequence_number","sequence_id","sequence_range","candidate_target_range","candidate_target_strand","candidate_target_sequence",@cols[2..$#cols]));
  while(<SD>){
    @cols=split(/\t/);
    $targetname=$cols[1];
    ($ok,$rgn_no,$seqrange,$target_range,$target_strand,$target_refpoint_5p_offset,$target_sequence)=RecoverTargetInfo($targetname); 
    if($ok==0){
       $copy_successful=0;
       last;
    }
    print SDC (join"\t",($rgn_no+1,$ID[$rgn_no],$seqrange,$target_range,$target_strand,$target_sequence,@cols[2..$#cols]));
  }
  close SD;
  close SDC;
  return $copy_successful;
}

# recovers target info from temporary name of candidate target site encoded in CasValue input file.

sub RecoverTargetInfo{
   my $targetname=$_[0];
   my $rgn_no="";
   my $target_5p_position;
   my $target_strand;
   my $target_range="";
   my $ok=1;  # return code
   my $seqchr;
   my $seqrange="";
   my $target_refpoint_5p_offset="";
   my $target_sequence="";
   my $tgtpam_lnth_m1=$CURR_CAS9_TARGETLEN+$CURR_CAS9_PAMLNTH-1;  
   if($targetname=~/^R(\d+)-(\d+)(r?)$/){
      $rgn_no=$1;
      $target_5p_position=$2;
      $target_strand=$3;
    }
    else {
      $ok=0;
      return ($ok,$rgn_no,$seqrange,$target_range,$target_strand,$target_refpoint_5p_offset,$target_sequence);
    }
    if($target_strand eq"r"){
       $target_strand="-";
    }
    elsif($target_strand eq ""){
       $target_strand="+";
    }
    else{
      $ok=0;
      return ($ok,$rgn_no,$seqrange,$target_range,$target_strand,$target_refpoint_5p_offset,$target_sequence);
    }
    if($EXTRACT[$rgn_no]==1){
       $seqchr=$CHR[$rgn_no];
    }
    else{
       $seqchr="FASTA";
    }
    if($target_strand eq "+"){
      $target_range="$seqchr:$target_5p_position-".($target_5p_position+$tgtpam_lnth_m1);
    }
    else {
      $target_range="$seqchr:".($target_5p_position-$tgtpam_lnth_m1)."-$target_5p_position";
    }
    $seqrange="$seqchr:$START[$rgn_no]-$END[$rgn_no]";
    ($target_refpoint_5p_offset,$target_sequence)=split/\t/,$TGT_CANDIDATES{$targetname};
    return ($ok,$rgn_no,$seqrange,$target_range,$target_strand,$target_refpoint_5p_offset,$target_sequence);
}

# FLOWCHART NOTE:
# The following subroutine performs the logic in the box "Report accepted Cas9 targets"
sub ReportTargets{
  my $CasValue_output_datafile=$_[0];
  my $i;
  my $rgn_no;
  my $target_refpoint_5p_offset;
  my $target_strand;
  my $target_sequence;
  my $outdatafile="$OUTPREF.data.txt";
  my $outstatfile="$OUTPREF.stat.txt";
  my $hdr;
  my $targetname;
  my $threshold_rejection_score;
  my $all_matches_analyzed;
  my $non_standard_base_matches;
  my $reported_targets=0;
  my $target_range;
  my %sequence_reported_targets=();
  my $num_sequences_with_targets;
  my $seqrange;
  my $ok; 
  my @co=(); 
  
  Msg("\nAnalyzing $CASVALUE_PGMNAME output and reporting targets to $outdatafile");

  open(O,">$outdatafile");
  print O "target_number\tsequence_number\tsequence_id\tsequence_range\tsequence_refpoint\tsequence_acceptnum".
          "\tsequence_other_annotation\ttarget_threshold_rejection_score\ttarget_all_matches_analyzed?\ttarget_nonstandard_basecode_matches".
          "\ttarget_location\ttarget_strand\ttarget_refpoint_5p_offset\ttarget_sequence\n";
  open(CO,$CasValue_output_datafile);
  $hdr=<CO>;
  while(<CO>){
    s/\s+$//;
    @co=split(/\t/); 
    $targetname=$co[1]; 
    $threshold_rejection_score=$co[2]; 
    $all_matches_analyzed=$co[3]; 
    $non_standard_base_matches=$co[4]; 
    ($ok,$rgn_no,$seqrange,$target_range,$target_strand,$target_refpoint_5p_offset,$target_sequence)=RecoverTargetInfo($targetname); 
    unless($ok==1){
      Msg("Error: Unanalyzable $CASVALUE_PGMNAME output line: $_");
      MsgContinue("$CASVALUE_PGMNAME output datafile kept as $CasValue_output_datafile.");
      MsgContinue("Program terminating.");
      close CO;
      close O;
      close I;
      exit;
    }
    $sequence_reported_targets{$rgn_no}++;
    $reported_targets++;
    print O "$reported_targets\t".($rgn_no+1)."\t$ID[$rgn_no]\t$seqrange\t$REFPOINT[$rgn_no]\t".
            ($CASVALUE_A eq ""?$ACCEPTNUM[$rgn_no]:$CASVALUE_A).  
            "\t$OTHER_ANNOTATION[$rgn_no]\t$threshold_rejection_score\t$all_matches_analyzed\t$non_standard_base_matches".
            "\t$target_range\t$target_strand\t$target_refpoint_5p_offset\t$target_sequence\n";
  }
  close O;
  close CO;
  
  if($DEBUG==0){
    unlink $CasValue_output_datafile;
  }
  
  Msg("\nReporting target statistics by sequence in $outstatfile.");
  open(OS,">$outstatfile");
  print OS "sequence_number\tsequence_id\tsequence_range\tnumber_reported_targets\n";
  for($i=0;$i<@ID;++$i){
    if($EXTRACT[$i]==1){
       $seqchr=$CHR[$i];
    }
    else{
       $seqchr="FASTA";
    }
    $seqrange="$seqchr:$START[$i]-$END[$i]";
    print OS ($i+1);
    print OS "\t$ID[$i]\t$seqrange\t".($sequence_reported_targets{$i}eq""?0:$sequence_reported_targets{$i})."\n";
  }    
  close OS;
  $num_sequences_with_targets=scalar (keys %sequence_reported_targets);
  Msg("\n$reported_targets targets reported for $num_sequences_with_targets out of ".(scalar @ID)." input sequences.\n");

}

sub Array{
  my @a=();
  return \@a;
}        

sub Hash{
  my %h=();
  return \%h;
}

sub RC{
  my $s=$_[0];
  my @s=split//,$s; 
  my $rc=join("",(reverse @s)); 
  $rc=~tr/ACGT/TGCA/;
  return $rc;
}


sub TimeStamp {
  my $large = $_[0];
  my $ts = "";
  my $yr = (localtime(time()))[5]+1900;
  my $mo = (localtime(time()))[4];
  my $da = (localtime(time()))[3];
  my $hr = (localtime(time()))[2];
  my $mi = (localtime(time()))[1];
  my $se = (localtime(time()))[0];

  if ($large eq "LARGE") {   
    $mo = (Jan,Feb,Mar,Apr,May,Jun,Jul,Aug,Sep,Oct,Nov,Dev)[$mo];
    $ts = "$mo $da, $yr at ";
  }
  if ($hr<10){$hr = "0$hr";} 
  if ($mi<10){$mi = "0$mi";} 
  if ($se<10){$se = "0$se";} 
  return "$ts$hr:$mi:$se";
}

sub Msg{
  my $msg=$_[0];
  if($msg=~/^(\n+)/){
     print I $1;
     $msg=~s/^\n+//;
  }
  print I TimeStamp().": $msg\n";
}
sub MsgContinue{
  my $msg;
  foreach $msg(@_){
    print I " "x(length(TimeStamp())+2);
    print I "$msg\n";
  }
}

# FLOWCHART NOTE:
# The following subroutine performs the logic in the box "Analyze parameters & retrieve config file info"
sub AnalyzeParameters{
  use IO::Handle;
  my $i;
  my $vol;  # Used to get the pgmname of the CASVALUE_PROGRAM
  my $dir;  # Used to get the pgmname of the CASVALUE_PROGRAM

  $CASVALUE_PARAMETER_STRING=""; 
  $CASVALUE_PROGRAM_PARAMETER=""; 
  $INFILE="";
  $OUTPREF="";
  $TFILTER="";
  $TFILTER_DFLT=5;
  $CONFIG_FILE="";
  $CONFIG_CAS9_FILE="";
  $CONFIG_FILE_DFLT="CASFINDER_CONFIG.txt";   
  $CONFIG_CAS9_FILE_DFLT="CASFINDER_CAS9_CONFIG.txt"; 
  $CAS9="";
  $CAS9_DFLT="SP";
  $GENOME="";
  $GENOME_DFLT="hg19";
  $FORCE_SEARCHREPEAT="";
  $CASVALUE_K=""; 
  $CASVALUE_X=""; 
  $CASVALUE_D=""; 
  $CASVALUE_A=""; 
  $PGMDOCUMENTATION="http://arep.med.harvard.edu/CasFinder"; 
  $DEBUG=0;
  
  for($i=0;$i<@_;++$i){
     if($_[$i] eq"-i"){
        if($i==$#_){
           die "No value specified for parameter $_[$i]\n";
        }
        $INFILE=$_[$i+1];
        ++$i;
     }
     elsif($_[$i] eq"-o"){
        if($i==$#_){
           die "No value specified for parameter $_[$i]\n";
        }
        $OUTPREF=$_[$i+1];
        ++$i;
     }
     elsif($_[$i] eq"-t"){
        if($i==$#_){
           die "No value specified for parameter $_[$i]\n";
        }
        $TFILTER=$_[$i+1];
        unless($TFILTER=~/^\d+/ && $TFILTER>0){
           die "Invalid value specified for $_[$i] parameter.\n";
        }
        ++$i;
     }
     elsif($_[$i] eq"-g"){
        if($i==$#_){
           die "No value specified for parameter $_[$i]\n";
        }
        $GENOME=$_[$i+1];
        ++$i;
     }
     elsif($_[$i] eq"-c"){  
        if($i==$#_){
           die "No value specified for parameter $_[$i]\n";
        }
        $CAS9=$_[$i+1];
        ++$i;
     }
     elsif($_[$i] eq"-k"){ 
        if($i==$#_){
           die "No value specified for parameter $_[$i]\n";
        }
        $CASVALUE_K=$_[$i+1];
        ++$i;
     }     
     elsif($_[$i] eq"-A"){ 
        if($i==$#_){
           die "No value specified for parameter $_[$i]\n";
        }
        $CASVALUE_A=$_[$i+1];
        ++$i;
     }     
     elsif($_[$i] eq"-x"){ 
        if($i==$#_){
           die "No value specified for parameter $_[$i]\n";
        }
        $CASVALUE_X=$_[$i+1];
        ++$i;
     }     
     elsif($_[$i] eq"-CONFIG"){
        if($i==$#_){
           die "No value specified for parameter $_[$i]\n";
        }
        $CONFIG_FILE=$_[$i+1];
        ++$i;
     }
     elsif($_[$i] eq"-CONFIG_CAS9"){
        if($i==$#_){
           die "No value specified for parameter $_[$i]\n";
        }
        $CONFIG_CAS9_FILE=$_[$i+1];
        ++$i;
     }     
     elsif($_[$i] eq"-CASVALUE_PROGRAM"){ 
        if($i==$#_){
           die "No value specified for parameter $_[$i]\n";
        }
        $CASVALUE_PROGRAM_PARAMETER=$_[$i+1];
        ++$i;
     }          elsif($_[$i]=~/^-CV$/){
        $CASVALUE_PARAMETER_STRING=$_[$i+1];
        ++$i;
     }
     elsif($_[$i] eq"-R"){
        $FORCE_SEARCHREPEAT=1;
     }
     elsif($_[$i] eq"-D"){
        $CASVALUE_D=1;
     }
     elsif($_[$i] eq"-DEBUG"){
        $DEBUG=1;
     }
     elsif($_[$i]=~/^-\?/ || $_[$i]=~/^-h(e|el|elp)?$/i){
        print "$PROGRAM program options:\n";
        print " -CONFIG: CASVALUE_PROGRAM configuration file (default=$CONFIG_FILE_DFLT)\n";
        print " -CONFIG_CAS9: CASVALUE_PROGRAM Cas9 configuration file (default=$CONFIG_CAS9_FILE_DFLT)\n";
        print " -i: input file (required)\n";
        print " -o: output file prefix (required)\n";
        print " -g: target genome (default=$GENOME_DFLT)\n";
        print " -c: Cas9 for which targets will be searched (default=$CAS9_DFLT)\n";
        print " -t: Filter out T-strings with this or greater length (default=$TFILTER_DFLT).\n";
        print " -k: CASVALUE_PROGRAM -k parameter (optional)\n";  
        print " -x: CASVALUE_PROGRAM -x parameter (optional)\n";  
        print " -CV: other CASVALUE_PROGRAM program parameter default overrides (optional)\n";
        print " -R: Force search of repeat regions in input sequences (optional)\n";
        print " -A: CASVALUE_PROGRAM -A parameter (optional)\n"; 
        print " -D: CASVALUE_PROGRAM -D parameter (optional)\n"; 
        print " -CASVALUE_PROGRAM: override CONFIG file specification of CasValue program name (optional)\n";
        print "\nInput file summary:\n";
        print " (1) The -i input file may contain a combination of ordinary FASTA sequences and FASTA headers\n";
        print "   provided without sequence that contain the keyword \"*EXTRACT*\", which indicate\n";
        print "   sequences to be extracted from the -g genome.  The FASTA header of such sequences must\n";
        print "   contain a range specification \"range=\\S+:\\d+-\\d+\", where the \\S+ specifies a\n"; 
        print "   chromosome, the first \\d+ indicates a start position, and the second \\d+ indicates\n";
        print "   an end position.\n";
        print " (2) \"refpoint=\\d+\" and \"reforient=(+|-)\" keywords may be provided with either kind\n";
        print "   of sequence specification. These instruct $PROGRAM to report the location of the 5' end\n";
        print "   of any accepted Cas9 target as an offset from a user-specified position within the sequence.\n";
        print "   The refpoint location must be in the range interval, where, for extracted sequences, the\n";
        print "   range is specified as in NOTE (1). For ordinary (non-extracted) FASTA sequences, the range\n";
        print "   is assumed to be 1-L where L is the length of provided sequence.\n";
        print " (3) Either kind of sequence may contain an \"accept=\\d+\" keyword. This tells CASVALUE_PROGRAM\n";
        print "   to accept any candidate Cas9 target site within the sequence that contains no more than\n";
        print "   the indicated number of matches within the -g genome whose match scores are within its\n";
        print "   scoring thresholds.\n";
        print " (4) By default $PROGRAM will not search for targets in repetitive sequence, where this is\n";
        print "   indicated by lower case a,c,g, and t basecodes in the sequence.  This behavior may be\n";
        print "   overridden for a given sequence by specifying \"searchrepeat=Y\" in its FASTA header.\n";
        print "   However, the -R parameter may be used to override individual sequence-specific \"searchrepeat\"\n";
        print "   indications and force all repetitive sequence regions to be searched for targets.\n";
        print " (5) The default CASVALUE_PROGRAM file is in the CONFIG file and may be overridden by the\n";
        print "   -CASVALUE_PROGRAM parameter. The program file should be fully qualified and include its\n";
        print "   directory. Default CASVALUE_PROGRAM scoring thresholds and other parameters can be displayed\n";
        print "   with the command  \"perl CASVALUE_PROGRAM -h\".  Default CASVALUE_PROGRAM parameters can\n";
        print "   be overriden by entering a CASVALUE_PROGRAM parameter string with the $PROGRAM -CV parameter.\n\n";
        print "\nDetailed documentation on program options and input file specifications can be found on:\n";
        print "$PGMDOCUMENTATION.\n";
       exit;
     }
     else{
        die "Unrecognized parameter $_[$i]\n";
     }
  }
  if($TFILTER eq""){
    $TFILTER=$TFILTER_DFLT;
  }
  if($CONFIG_FILE eq""){
    $CONFIG_FILE=$CONFIG_FILE_DFLT;
  }
  if($CONFIG_CAS9_FILE eq""){
    $CONFIG_CAS9_FILE=$CONFIG_CAS9_FILE_DFLT;
  }  
  if($CAS9 eq ""){
     $CAS9=$CAS9_DFLT;
  }
  if($GENOME eq ""){
     $GENOME=$GENOME_DFLT;
  }
  if($INFILE eq ""){
     die "No input file specified (-i)\n";
  }
  if($OUTPREF eq ""){
     die "No output file prefix specified (-o)\n";
  }
  if(! -e $INFILE){
     die "Specified input file -i does not exist.\n";
  }
  if($CASVALUE_PROGRAM_PARAMETER ne "" && ! -e $CASVALUE_PROGRAM_PARAMETER){ 
     die "Specified -CASVALUE_PROGRAM CasValue program file override does not exist.\n";
  }
  open(I,">$OUTPREF.info.txt");
  autoflush I 1;
  
  print I "Parameters for execution of $PROGRAM on ".TimeStamp("LARGE").":\n";
  print I "-CONFIG $CONFIG_FILE (general configuration file)\n";
  print I "-CONFIG_CAS9: $CONFIG_CAS9_FILE (Cas9 configuration file)\n";
  print I "-i:   $INFILE (input sequence specification file)\n";
  print I "-o:   $OUTPREF (output file prefix)\n";
  print I "-t:   $TFILTER (length at which T-strings will be filtered)\n";
  print I "-c:   $CAS9 (target Cas9 protein)\n";
  print I "-g:   $GENOME (target genome for sequence extraction and specificity analysis)\n";
  if($CASVALUE_PROGRAM_PARAMETER ne ""){
    print I "-CASVALUE_PROGRAM:  $CASVALUE_PROGRAM_PARAMETER (CasValue program name override)\n";
  }
  if($CASVALUE_K ne ""){  
    print I "-k:   $CASVALUE_K (CasValue program -k parameter)\n";
  }
  if($CASVALUE_X ne ""){  
    print I "-x:   $CASVALUE_X (CasValue program -x parameter)\n";
  }
  print I "-CV:  $CASVALUE_PARAMETER_STRING ($CASVALUE_PROGRAM program parameter overrides)\n";
  if($FORCE_SEARCHREPEAT==1){
    print I "-R:   Force search of all repetitive regions in input sequences.\n";
  }
  if($CASVALUE_A ne ""){  
    print I "-A:   $CASVALUE_A (CasValue program -A parameter)\n";
  }
  if($CASVALUE_D==1){  
    print I "-D:   Specify -D option to CasValue program.\n";
  }
  print I "\n";

  if($DEBUG==1){
    my $dbgfile="$PROGRAM.DEBUG.out";
    open(DEBUG,">$dbgfile");
    Msg("\nProgram is in DEBUG mode.\n");
  }

  # Now read and integrate wtih configuration files
  CasFinderConfigFile->Read($CONFIG_FILE);     
  CasFinderConfigCas9File->Read($CONFIG_CAS9_FILE); 
 
  $CURR_GENOME_CHROMOSOME_ORDER=CasFinderConfigFile->Get("GENOME_CHROMOSOME_ORDER",$GENOME);
  $CURR_GENOME_CHROMOSOME_FILES=CasFinderConfigFile->Get("GENOME_CHROMOSOME_FILES",$GENOME);
  $CASVALUE_PROGRAM=CasFinderConfigFile->Get("CASVALUE_PROGRAM");
 
  $TEMPPREF=CasFinderConfigFile->Get("TEMPPREF");
  if($DEBUG==1){  # if in debug mode, reset temporary prefix
    $TEMPPREF="testtemp";
  }

  if($CURR_GENOME_CHROMOSOME_ORDER eq"" || $CURR_GENOME_CHROMOSOME_FILES eq""){
    print I "\nSpecified genome $GENOME not specified in configuration file $CONFIG_FILE\n";
    print I "Program terminating.\n";
    exit;
  }
 
  # check for existence of chromsome files

  foreach $chrfile(@$CURR_GENOME_CHROMOSOME_ORDER){
    if(! -e $$CURR_GENOME_CHROMOSOME_FILES{$chrfile}){ 
      print I "\nSpecified chromosome file for $chrfile for $GENOME in configuration file $CONFIG_FILE does not exist:\n$$CURR_GENOME_CHROMOSOME_FILES{$chrfile}\n";
      print I "Program terminating.\n";
      exit;
    }
  }
  Msg("System parameters read from configuration file $CONFIG_FILE:");
  MsgContinue("CasValue program = $CASVALUE_PROGRAM");
  MsgContinue("Temporary file prefix = $TEMPPREF\n");
  Msg("Chromosome files for genome $GENOME have been read from configuration file $CONFIG_FILE\n");

  if($CASVALUE_PROGRAM_PARAMETER ne ""){  
     $CASVALUE_PROGRAM=$CASVALUE_PROGRAM_PARAMETER;
     Msg("CasValue program name overridden to $CASVALUE_PROGRAM");
  }
  
  # Get just the program name of the CASVALUE_PROGRAM
  ($vol,$dir,$CASVALUE_PGMNAME)=File::Spec->splitpath($CASVALUE_PROGRAM);

  $CURR_CAS9_TARGETLEN=CasFinderConfigCas9File->Get("CAS9_TARGETLEN",$CAS9);
  $CURR_CAS9_PAM=CasFinderConfigCas9File->Get("CAS9_PAM",$CAS9);
  $CURR_CAS9_PAMLNTH=CasFinderConfigCas9File->Get("CAS9_PAMLNTH",$CAS9);  
  $CURR_CAS9_PAM_REGEXP=CasFinderConfigCas9File->Get("CAS9_PAM_REGEXP",$CAS9);

  if($CURR_CAS9_TARGETLEN eq""||$CURR_CAS9_PAM eq""||$CURR_CAS9_PAM_REGEXP eq""){
    print I "Specified Cas9 $CAS9 not defined in configuration file $CONFIG_CAS9_FILE\n";
    print I "Program terminating.\n";
    exit;
  }
  
  Msg("\nCas9 $CAS9 characteristics read from configuration file $CONFIG_CAS9_FILE:");
  MsgContinue("target_length=$CURR_CAS9_TARGETLEN,PAM=$CURR_CAS9_PAM ($CURR_CAS9_PAM_REGEXP),PAMLNTH=$CURR_CAS9_PAMLNTH\n");  

}
 
