# CasFinderConfigCas9File.pm
#
# perl package for reading the CasFinder system CONFIG_CAS9 file.  This package
# is used by CasFinder system programs CasFinder.pl and CasValue_v2.pl to 
# interrogate information on the file. See documentation on 
# http://arep.med.harvard.edu/CasFinder for details.
#
# AVAILABILITY AND CITATION
# 
# The CasFinder system of programs comprises CasFinder.pl, CasValue_v2.pl,
# the two perl packages CasFinderConfigFile.pm and CasFinderConfigCas9File.pm, 
# and the two configuration files CASFINDER_CONFIG.txt and 
# CASFINDER_CONFIG_CAS9.txt.
# 
# The CasFinder system of programs is being made available under the MIT 
# Licence (MIT).   
#
# Please cite the article: 
#
#   Aach J., Mali P., Church G.M..
#     CasFinder: Flexible algorithm for identifying specific Cas9 targets 
#     in genomes.  bioRxiv (doi)  
#
# when using, copying, or employing code from this system.
#
# COPYRIGHT NOTICE
# 
# The MIT License (MIT)
#
# Copyright (c) 2014 John Aach and the President and Fellows of Harvard University
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# CasFinderConfigFile.pm program CHANGE LOG:
# Created: 09/09/2013-05/12/2014 by John Aach
#
 package CasFinderConfigCas9File;
 require Exporter;

 our @ISA = qw(Exporter);
 our @EXPORT = qw(Read,
                  Get);
 return 1;

# Read Cas9 configuration file into storage
sub Read{
  my $pkg=$_[0];
  my $config_cas9_file=$_[1];
  my $ch;
  my $cas9="";
  my @pam;
  my $pam; 
  my @pam_re; 
  my $pam_re;
  my $pamlnth; 
  my @cfg=();  
  %CAS9_TARGETLEN=();
  %CAS9_SEED_START=();
  %CAS9_SEED_END=();
  %CAS9_PAM=();
  %CAS9_PAM_REGEXP=();
  %CAS9_PAM2=();
  %CAS9_PAM2_REGEXP=();
  %CAS9_PAMLNTH=(); # length of PAM (all PAMs and PAM2s must have same length, padded with Ns if needed).
  
  if($config_cas9_file eq ""){
     die "Error detected by $pkg: No configuration file name has been received by the package.\n";
  }   
  if(! -e $config_cas9_file){
     die "Error detected by $pkg: configuration file $config_cas9_file does not exist.\n";
  }
  open (C,$config_cas9_file);
  while(<C>){
    if(/^\s+$/){next;} # skip blank lines
    s/^\s+//; # get rid of initial whitespace
    if(/^#/){next;} # skip comments
    s/\s+$//;
    @cfg=split(/\s+/);
    if((scalar @cfg)!=5 && (scalar @cfg)!=6){
      die "Error: $pgk::Read(): Cas9 configuration file $config_cas9_file: line contains other than five or six columns:\n$_\n";
    }
    $cas9=$cfg[0];
    if($CAS9_TARGETLEN{$cas9} ne ""){
      die "Error: $pgk::Read(): Cas9 configuration file $config_cas9_file: duplicate definitions for Cas9 $cas9\n";
    }
    if($cfg[1]=~/^\d+$/&&$cfg[1]>0){
      $CAS9_TARGETLEN{$cas9}=$cfg[1];
    }
    else{
      die "Error: $pgk::Read(): Cas9 configuration file $config_cas9_file: $cas9 Invalid target length $cfg[1]\n";
    }
    if($cfg[2]=~/^\d+$/&&$cfg[2]>0&&$$cfg[2]<=$CAS9_TARGETLEN{$cas9}){
      $CAS9_SEED_START{$cas9}=$cfg[2];
    }
    else{
      die "Error: $pgk::Read(): Cas9 configuration file $config_cas9_file: $cas9 Invalid seed_start position $cfg[2]\n";
    }
    if($cfg[3]=~/^\d+$/&&$cfg[3]>0&&$$cfg[3]<=$CAS9_TARGETLEN{$cas9}&&$cfg[3]>$CAS9_SEED_START{$cas9}){
      $CAS9_SEED_END{$cas9}=$cfg[3];
    }
    else{
      die "Error: $pgk::Read(): Cas9 configuration file $config_cas9_file: $cas9 Invalid seed_end position $cfg[3]\n";
    }

    $CAS9_PAM{$cas9}=$cfg[4];
    $CAS9_PAM2{$cas9}=$cfg[5];
    # PAM array logic. PAM array is comma-separated list of PAMs that must all be same length
    @pam=split/,/,$CAS9_PAM{$cas9};
    $pamlnth=0;
    foreach $pam(@pam){
       if(length($pam)==0){
          die "Error: $pgk::Read(): Cas9 configuration file $config_cas9_file: $cas9 empty PAM in PAM array: $CAS9_PAM{$cas9}\n";
       }
       if($pamlnth==0){
          $pamlnth=length($pam);
       }
       elsif(length($pam)!=$pamlnth){
          die "Error: $pgk::Read(): Cas9 configuration file $config_cas9_file: $cas9 PAMs in PAM array have unequal lengths: $CAS9_PAM{$cas9}\n";
       }
    }
    $CAS9_PAMLNTH{$cas9}=$pamlnth;      
    # repeat checks with alternate PAM array and check that lengths == that of primary PAM(s)
    @pam=split/,/,$CAS9_PAM2{$cas9};
    foreach $pam(@pam){
       if(length($pam)==0){
          die "Error: $pgk::Read(): Cas9 configuration file $config_cas9_file: $cas9 empty PAM2 in PAM2 array: $CAS9_PAM2{$cas9}\n";
       }
       elsif(length($pam)!=$pamlnth){
          die "Error: $pgk::Read(): Cas9 configuration file $config_cas9_file: $cas9 PAMs in PAM2 array have lengths unequal to primary PAM length: $CAS9_PAM2{$cas9}\n";
       }
    }
    $CAS9_PAM_REGEXP{$cas9}=PAM2REGEXP($CAS9_PAM{$cas9},$cas9,$config_cas9_file);
    if($CAS9_PAM2{$cas9}ne""){
      $CAS9_PAM2_REGEXP{$cas9}=PAM2REGEXP($CAS9_PAM2{$cas9},$cas9,$config_cas9_file);
    }
  }
  close C;
}

sub Get{
  my $pkg=$_[0];
  my $var=uc $_[1];
  my $key=$_[2];
  if($var eq "CAS9_TARGETLEN"){
    if($CAS9_TARGETLEN{$key} eq ""){
      die "Error: $pkg Get(): No value defined for key $key in $pkg variable $var\n"; 
    }
    else {
      return $CAS9_TARGETLEN{$key};
    }
  }
  elsif($var eq "CAS9_SEED_START"){
    if($CAS9_SEED_START{$key} eq ""){
      die "Error: $pkg Get(): No value defined for key $key in $pkg variable $var\n"; 
    }
    else {
      return $CAS9_SEED_START{$key};
    }
  }
  elsif($var eq "CAS9_SEED_END"){
    if($CAS9_SEED_END{$key} eq ""){
      die "Error: $pkg Get(): No value defined for key $key in $pkg variable $var\n"; 
    }
    else {
      return $CAS9_SEED_END{$key};
    }
  }
  elsif($var eq "CAS9_PAM"){
    if($CAS9_PAM{$key} eq ""){
      die "Error: $pkg Get(): No value defined for key $key in $pkg variable $var\n"; 
    }
    else {
      return $CAS9_PAM{$key};
    }
  }
  elsif($var eq "CAS9_PAMLNTH"){
    if($CAS9_PAMLNTH{$key} eq ""){
      die "Error: $pkg Get(): No value defined for key $key in $pkg variable $var\n"; 
    }
    else {
      return $CAS9_PAMLNTH{$key};
    }
  }
  elsif($var eq "CAS9_PAM_REGEXP"){
    if($CAS9_PAM_REGEXP{$key} eq ""){
      die "Error: $pkg Get(): No value defined for key $key in $pkg variable $var\n"; 
    }
    else {
      return $CAS9_PAM_REGEXP{$key};
    }
  }
  elsif($var eq "CAS9_PAM2"){
    if($CAS9_PAM2{$key} eq ""){
      die "Error: $pkg Get(): No value defined for key $key in $pkg variable $var\n"; 
    }
    else {
      return $CAS9_PAM2{$key};
    }
  }
  elsif($var eq "CAS9_PAM2_REGEXP"){
    if($CAS9_PAM2_REGEXP{$key} eq ""){
      die "Error: $pkg Get(): No value defined for key $key in $pkg variable $var\n"; 
    }
    else {
      return $CAS9_PAM2_REGEXP{$key};
    }
  }
  else{
    die "Error: $pkg Get(): $pkg variable $var not defined.\n"; 
  }
}

# This routine tranlsates comma-separated string of PAMs including degenerate bases into a
# "|"-separated string of perl regexps that can be used in a pattern match.
sub PAM2REGEXP{
  my $pamarray=$_[0];
  my $cas9=$_[1];
  my $config_cas9_file=$_[2];
  my @pamarray=split/,/,$pamarray;
  my $pam_in_array;
  my @pamarray_regexp=();
  my @pam;
  my $return_regexp;
  my $ch;
  my $regexp;

  foreach $pam_in_array (@pamarray) { 
    @pam=split//,$pam_in_array;
    $regexp="";
    foreach $ch(@pam){
       if($ch=~/[ACGT]/i){
          $regexp.=$ch;
       }
       elsif($ch=~/B/i){
          $regexp.="[CGT]";
       }
       elsif($ch=~/D/i){
          $regexp.="[AGT]";
       }
       elsif($ch=~/H/i){
          $regexp.="[ACT]";
       }
       elsif($ch=~/V/i){
          $regexp.="[ACG]";
       }
       elsif($ch=~/R/i){
          $regexp.="[AG]";
       }
       elsif($ch=~/Y/i){
          $regexp.="[CT]";
       }
       elsif($ch=~/K/i){
          $regexp.="[GT]";
       }
       elsif($ch=~/M/i){
          $regexp.="[AC]";
       }
       elsif($ch=~/S/i){
          $regexp.="[CG]";
       }
       elsif($ch=~/W/i){
          $regexp.="[AT]";
       }
       elsif($ch=~/N/i){
          $regexp.="[ACGT]";
       }
       else {
         die "Error in Cas9 configuration file $config_cas9_file: $cas9 PAM $pam contains invalid base code $ch\n";
       }
    }
    push @pamarray_regexp,$regexp;
  }
  if((scalar @pamarray_regexp)==1){
    $return_regexp=$pamarray_regexp[0];
  }
  else {
    $return_regexp="(".(join"|",@pamarray_regexp).")"; 
  }
  return $return_regexp;
}
# This routine takes the reverse complement of a "|"-separated string of PAM regexps of the 
# sort generated by PAM2REGEXP
sub RCREGEXP{
  my $pkg=$_[0];
  my $regexp_string=$_[1];
  my $rest;
  my @regexp_array;
  my $regexp;
  my $rest;
  my $group;
  my $letters;
  my $rcregexp;
  my @rcregexp_array=();
  my $return_rcregexp;  
  
  $regexp_string=~s/^\(//;
  $regexp_string=~s/\)$//;
  @regexp_array=split/\|/,$regexp_string; 

  foreach $regexp(@regexp_array){
    $rcregexp="";
    $rest=$regexp;
    while($rest ne ""){
      if($rest=~/^([ACGT])(.*)$/){
         $letters=$1;
         $rest=$2;
         $group=0;
      }
      elsif($rest=~/^\[([ACGT]+)\](.*)$/){
         $letters=$1;
         $rest=$2;
         $group=1;
      }
      else{
         die "Error: $pkg RCREGEXP: Invalid parse of $rest in $regexp within $regexp_array\n";  
      }
      $letters=~tr/ACGT/TGCA/;
      my @letters=split//,$letters; 
      $letters=(join"",sort @letters);
      if($group==0){
         $rcregexp=$letters.$rcregexp;
      }
      else{
        $rcregexp="[".$letters."]".$rcregexp;
      }
    }
    push @rcregexp_array,$rcregexp; 
  }
  if((scalar @rcregexp_array)==1){
    $return_rcregexp=$rcregexp_array[0];
  }
  else {
    $return_rcregexp="(".(join"|",@rcregexp_array).")"; 
  }
  return $return_rcregexp; 
}