# CasFinderConfigFile.pm
#
# perl package for reading the CasFinder system CONFIG file.  This package
# is used by CasFinder system programs CasFinder.pl and CasValue_v2.pl to 
# interrogate information on the file. See documentation on 
# http://arep.med.harvard.edu/CasFinder for details.
#
# AVAILABILITY AND CITATION
# 
# The CasFinder system of programs comprises CasFinder.pl, CasValue_v2.pl,
# the two perl packages CasFinderConfigFile.pm and CasFinderConfigCas9File.pm, 
# and the two configuration files CASFINDER_CONFIG.txt and 
# CASFINDER_CONFIG_CAS9.txt.
# 
# The CasFinder system of programs is being made available under the MIT 
# Licence (MIT).   
#
# Please cite the article: 
#
#   Aach J., Mali P., Church G.M..
#     CasFinder: Flexible algorithm for identifying specific Cas9 targets 
#     in genomes.  bioRxiv (doi)   
#
# when using, copying, or employing code from this system.
#
# COPYRIGHT NOTICE
# 
# The MIT License (MIT)
#
# Copyright (c) 2014 John Aach and the President and Fellows of Harvard University
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# CasFinderConfigFile.pm program CHANGE LOG:
# Created: 09/09/2013-05/12/2014 by John Aach
#
 package CasFinderConfigFile;
 require Exporter;
 use File::Spec;

 our @ISA = qw(Exporter);
 our @EXPORT = qw(Read 
                  Get);
 %GENOME_BOWTIE_INDEX=();
 %GENOME_CHROMOSOME_FILES=();
 %GENOME_CHROMOSOME_ORDER=();
 $BOWTIE_EXECUTABLE="";
 $CASVALUE_PROGRAM=""; 
 $TEMPPREF="";
 return 1;

# Read general configuration file into storage
sub Read{
  my $pkg=$_[0];
  my $config_file=$_[1];
  my $in_genome=0;
  my $curr_genome="";
  my $chromdir="";
  my $a; # array reference variable
  my $h; # hash reference variable
  my $i; 
  my $timeval=time();
  my $in_chrom=0; 
  my @cfg=();  
  %GENOME_BOWTIE_INDEX=();
  %GENOME_CHROMOSOME_FILES=();
  %GENOME_CHROMOSOME_ORDER=();
  
  if($config_file eq ""){
     die "Error detected by $pkg: No configuration file name has been received by the package.\n";
  }   
  if(! -e $config_file){
     die "Error detected by $pkg: configuration file $config_file does not exist.\n";
  }
  open (C,$config_file);
  while(<C>){
    if(/^\s+$/){next;} # skip blank lines
    s/^\s+//; # get rid of initial whitespace
    if(/^#/){next;} # skip comments
    s/\s+$//;
    @cfg=split(/\s+/);  
    if($cfg[0]=~/^bowtie_executable$/i){ $BOWTIE_EXECUTABLE=$cfg[1]; }
    elsif($cfg[0]=~/^casvalue_program$/i){ $CASVALUE_PROGRAM=$cfg[1]; } 
    elsif($cfg[0]=~/^temp_file_prefix$/i){ 
      $TEMPPREF=$cfg[1]; 
      $TEMPPREF=~s/PROGRAM/$pkg/g;
      $TEMPPREF=~s/TIME/$timeval/;
    }
    elsif($cfg[0]=~/^genome$/i){ 
      $curr_genome=$cfg[1];
      if($in_chrom==1){ 
         print STDERR "Warning: Possible error in configuration file $config_file: \"genome\" statement appears in a \"chromosomes\" section\n";
         print STDERR "\tA \"chromosomes\" section for a previous genome may be missing an \"end chromosomes\" statement.  Statement is\n$_\n";
      }
      if($curr_genome eq ""){
         die "Error in configuration file $config_file: No genome name specified in genome line\n";
      }
      $in_genome=1;
      if($GENOME_CHROMOSOME_FILES{$curr_genome} eq ""){
        $GENOME_CHROMOSOME_FILES{$curr_genome}=Hash();
      }
      else { 
         die "Error in configuration file $config_file: genome $cfg[1] specified multiple times.\n";
      }         
      if($GENOME_CHROMOSOME_ORDER{$curr_genome} eq ""){
        $GENOME_CHROMOSOME_ORDER{$curr_genome}=Array();
      }
    }
    elsif($cfg[0]=~/^bowtie_index$/i){
      if($in_genome==0){
        die "Error in configuration file $config_file: bowtie_index specified outside of genome section\n";
      }
      if($cfg[1] eq ""){
         die "Error in configuration file $config_file: No bowtie_index prefix specified in bowtie_index line\n";
      }
      $GENOME_BOWTIE_INDEX{$curr_genome}=$cfg[1];
    }
    elsif($cfg[0]=~/^chromosomes$/i){
      if($in_genome==0){
        die "Error in configuration file $config_file: chromosomes line specified outside of genome section\n";
      }
      if($cfg[1] eq ""){
         die "Error in configuration file $config_file: No chromosome file directory specified in chromosomes line\n";
      }
      $chromdir=$cfg[1];
      $in_chrom=1; 
    }
    elsif($cfg[0]=~/^end$/i && $cfg[1]=~/^chromosomes$/i){ 
       $in_chrom=0;
    }
    elsif($in_chrom==1){ # assume chromosome in current genome  
      if($in_genome==0){
         die "Error in configuration file $config_file: chromosome $cfg[0] specified outside of genome section\n";  
      }
      if($cfg[1] eq ""){
         die "Error in configuration file $config_file: No chromosome file name specified for chromosome $cfg[0] in genome $curr_genome\n";
      }
      $a=$GENOME_CHROMOSOME_ORDER{$curr_genome};
      push @$a,$cfg[0];
      $h=$GENOME_CHROMOSOME_FILES{$curr_genome};
      $$h{$cfg[0]}=File::Spec->catfile($chromdir,$cfg[1]);
    }
    else{ 
      die "Error in configuration file $config_file: Unrecognized statement:\n$_\n";
    }
  }
  close C;
  if($in_chrom==1){ 
     die "Error in configuration file $config_file: File ended in a chromosomes section. Missing an  \"end chromosomes\" statement\n";
  }
  # configuration file checks 
  if($TEMPPREF eq ""){
     die "Error in configuration file $config_file: No temp_file_prefix has been specified.\n";
  }
  if($BOWTIE_EXECUTABLE eq ""){
     die "Error in configuration file $config_file: No bowtie_executable has been specified.\n";
  }
  if(! -e $BOWTIE_EXECUTABLE){
     die "Error in configuration file $config_file: Specified bowtie_executable $BOWTIE_EXECUTABLE does not exist.\n";
  }
  if($CASVALUE_PROGRAM eq ""){
     die "Error in configuration file $config_file: No casvalue_program has been specified.\n";
  }
  if(! -e $CASVALUE_PROGRAM){
     die "Error in configuration file $config_file: Specified casvalue_program $CASVALUE_PROGRAM does not exist.\n";
  }
}
sub Get{
  my $pkg=$_[0];
  my $var=uc $_[1];
  my $key=$_[2];
  if($var eq "GENOME_BOWTIE_INDEX"){
    if($GENOME_BOWTIE_INDEX{$key} eq ""){
      die "Error: $pkg Get(): No value defined for key $key in $pkg variable $var\n"; 
    }
    else {
      return $GENOME_BOWTIE_INDEX{$key};
    }
  }
  elsif($var eq "GENOME_CHROMOSOME_FILES"){
    if($key eq ""){
      die "Error: $pkg Get() called without key for $pkg variable $var\n"; 
    }
    if($GENOME_BOWTIE_INDEX{$key} eq ""){
      die "Error: $pkg Get(): No value defined for key $key in $pkg variable $var\n"; 
    }
    else {
      return $GENOME_CHROMOSOME_FILES{$key};
    }
  }
  elsif($var eq "GENOME_CHROMOSOME_ORDER"){
    if($key eq ""){
      die "Error: $pkg Get() called without key for $pkg variable $var\n"; 
    }
    if($GENOME_CHROMOSOME_ORDER{$key} eq ""){
      die "Error: $pkg Get(): No value defined for key $key in $pkg variable $var\n"; 
    }
    else {
      return $GENOME_CHROMOSOME_ORDER{$key};
    }
  }
  elsif($var eq "BOWTIE_EXECUTABLE"){
    return $BOWTIE_EXECUTABLE;
  }
  elsif($var eq "CASVALUE_PROGRAM"){  
    return $CASVALUE_PROGRAM;
  }
  elsif($var eq "TEMPPREF"){
    return $TEMPPREF;
  }
  else{
    die "Error: $pkg Get(): $pkg variable $var not defined.\n"; 
  }
}
sub Hash{
  my %h=();
  return\%h;
}
sub Array{
  my @a=();
  return\@a;
}