# CasValue_v2.pl
#
# Perl script for evaluating the specificity of Cas9 target sequences in
# a genome.
#
# CasValue_v2.pl is a specific version of the CasValue program that
# is being distributed with the CasFinder system.  The purpose of the
# CasFinder system is to find Cas9 sites in input sequences provided
# or specified by a user that pass specificity checks for potential
# off-targets in a particular genome.  Within this system CasFinder.pl is
# responsible for identifying candidate Cas9 sites in the user-provided
# input, and CasFinder.pl calls CasValue_v2.pl in order to test the
# specificity of these candidate sites.  However, CasValue_v2.pl is
# designed so that it can be run independently of CasFinder.pl and so
# allow users to test the specificity of Cas9 sites determined by other
# programs or algorithms.  When called by CasFinder.pl, CasFinder.pl uses
# the same command line interface that may be used to run CasValue_v2.pl
# independently.
#
# A flow chart that presents the high-level logic of CasValue_v2 is 
# available as Supporting Info, Figure S2 of the article describing
# the CasFinder system (see citation below).  Within the code below
# the notation "FLOWCHART NOTE" signifies a call or block of code that
# is represented in the flow chart. Note that perl variable names in 
# CAPITAL LETTERS correspond to the data items in the box labeled 
# "CasValue internal memory" in the flow chart.
#
# CasFinder and the CasValue program both use a pair of configuration
# files, one of which defines genome and system files and resources,
# and the other of which defines the characteristics of the Cas9s that
# will be recognized by the programs.  However, neither program accesses
# these files directly but rather through a pair of perl packages that
# encapsulate access to them.  These packages are CasFinderConfigFile.pm
# and CasFinderConfigCas9File.pm.  CasValue_v2 gets information on the
# target sequence lengths and PAMs of defined Cas9s from this latter
# package, and information on the genome in which it is to check targets
# for specificity from the former package.
#
# Detailed instructions on how to install and use CasFinder.pl and
# CasValue_v2.pl, as well as other program documentation, can be found
# on http://arep.med.harvard.edu/CasFinder.  The following summarizes
# CasValue_v2.pl's program logic:
#
# 1. CasValue_v2.pl first analyzes its program options and reads
# information from the CasFinder system configuration files.  The program
# options will instruct it as to which Cas9 it is to process, which
# genome it is to search for potential off-targets, and also specify
# the input file.  Program options may also be present that override
# Cas9-specific information taken from the configuration files, e.g.,
# the position of the seed sequence within a full target sequence.
#
# 2. CasValue_v2.pl then reads the input file of target sequences, which
# are provided in FASTA form.  These sequences must contain *targeting
# sequences only* and not PAMs as well. In searching the genome it has been
# instructed to analyze, CasValue_v2 will find all sequences in the genome
# (within defined limits, see below) that match each targeting sequence
# and are followed by a recognized PAM for the Cas9.  (Thus, provision
# of a PAM in the input file with the targeting sequence is superfluous!)
#
# 3. To find matches to input targeting sequences in the specified
# genome, CasValue_v2.pl starts by looking for matches to the 'seed'
# region of the input sequence with up to a limited number of mismatches
# (specified by the -sm parameter).  It does this by formating a temporary
# FASTA file of seed regions from all of its input sequences and calling
# the bowtie program with a corresponding -v parameter to retrieve the
# genome matches.  Because seeds are short sequence regions and bowtie
# can return many matches, CasValue_v2.pl allows specification of the
# bowtie -k parameter to limit the number of returned matches per seed.
# This can result in incompletely analyzed input sequences (see below).
#
# 4. When bowtie execution has completed, CasValue_v2.pl goes through
# the bowtie output file to find the genome location of all returned
# matches, and then goes to chromosome sequences files for the genome and
# extracts the full Cas9 footprint containing the match -- where the full
# footprint conatins both the full targeting sequence at the match and
# the bases in the following PAM positions.  To make extraction of these
# footprints more efficient, CasValue_v2.pl batches all match locations by
# chromosomes and reads through the sequences one chromosome at a time,
# collecting and evaluating footprints as it goes. The evaluation of
# each footprint is as follows: If the footprint's PAM region does not
# correspond to a standard or 'alternate' PAM for the Cas9, the footprint
# is not considered. Otherwise, a weighted sum of the number of mismatches
# in the seed region, the number of mismatches in the non-seed region, plus
# an additional 'cost' factor (if the PAM region contains an 'alternate'
# vs. a primary PAM) is computed as a score.  The weights and alternate PAM
# cost factors are specified by program options or use program defaults,
# while the PAMs, targeting region lengths, and seed region locations,
# are specified in one of the configuration files.  For instance,
# in the configuration files distributed with the Casfinder system,
# for S. pyogenes Cas9, the target length is 20, the seed start is at
# position 8 and seed end is at position 20, while the default weights
# for both seed and non-seed mismatches are both 1, and the default cost
# of an alternate PAM (here NAG vs. the canonical NGG PAM) is also 1.
#
# 5. CasValue_v2.pl considers low-scoring Cas9 footprints for an
# input targeting sequence as potential off-targets for this sequence.
# Therefore, as CasValue_v2.pl goes through and scores all Cas9 footprints
# from seed matches in the genome, it keeps track of the lowest scores
# for all matches detected for each input target.  Each input sequence is
# considered to have an 'accept' number, which is either specified in its
# input FASTA header or otherwise specified through CasValue_v2.pl program
# options, and the number of lowest scores maintained for each input
# targeting sequence is one more than its 'accept' number.  Meanwhile, the
# CasValue_v2.pl program option -x defines the score threshold at or below
# which a Cas9 footprint is considered a potential Cas9 binding site when
# Cas9 is provided a crRNA or sgRNA with the input targeting sequence. If,
# after evaluating all scored matches for an input targeting sequence, the
# total number of matches with scores <= -x is <= the accept number, the
# target is considered 'accepted'.  This mechanism allows CasValue_v2.pl
# to check for targets that may be either endogenous or exogenous
# to the genome being checked.  For endogenous targets, the 'accept'
# number should be 1 to account for the endogenous target site itself,
# which will be found to have a score of 0.  For exogenous targets, the
# 'accept' number should be 0.  The CasValue_v2.pl default value of -x is
# 3, so that, given other defaults, input targeting sequences with 'accept'
# number 1 are rejected only if there is another site in the genome with <=
# 3 mismatches to the targeting sequence followed by a primary PAM, or 2
# mismatches followed by an alternate PAM.  Additionally, by keeping track
# of one more than the 'accept' number of lowest scores, CasValue_v2.pl
# can compute a 'threshold rejection score' for each input target that
# is accepted.  This value is the next lowest score above -x that it has
# detected in the genome for each accepted input targeting sequence in
# its evaluation of potential off- targets.  Higher 'threshold rejection
# scores' indicate greater distance in sequence space to the closest
# footprints in the genome that could be considered off-targets, within
# the limits of the search algorithm.  However, if the bowtie -k parameter
# is set to limit bowtie-returned seed matches in the genome, some input
# targeting sequences that CasValue_v2.pl accepts may be incorrectly
# evaluated, either by dint of having undetected additional matches
# with score <= -x (which would have caused rejection of the targeting
# sequence), or by dint of overestimating the 'threshold rejection score'.
# CasValue_v2.pl thus also maintains and reports an indicator for each
# accepted target, which flags any accepted input targeting sequence for
# which bowtie returned exactly -k matches.  Finally, CasValue_v2.pl
# also reports a count of matches that have non-standard base codes,
# since these (if any are returned) may represent matches in the genome
# for which the number of mismatches against the target, and thus the
# computed score for the match, are uncertain.
#
# 6. When its evaluation is complete, CasValue_v2.pl produces an output
# 'data' file of all accepted input targeting sequences, along with these
# sequences' threshold rejection scores and non- standard base code match
# counts.  Additionally, if the CasValue_v2.pl -D option was specified,
# CasValue_v2.pl also produces a 'site_detail' file that reports, for each
# input targeting sequence, the locations and scoring details of all Cas9
# footprints detected for the sequence with scores <= -x.  The sequences
# in the output 'data' file are in the order of these sequences in the
# input file (although only the accepted sequences are in the 'data' file),
# while the sequences in the 'site_detail' file are in no particular order:
# These are put out as CasValue_v2.pl finds them when it retrieves Cas9
# footprints chromosome by chromosome.  When processing is complete,
# CasValue_v2.pl deletes the bowtie output file, which can be extremely
# large. (The 'site_detail' file, if requested, can also be voluminous.)
#
# USAGE: Help messages summarizing CasFinder.pl program options may be
# viewed by executing
#
#    perl CasValue_v2.pl -h
#
# from a linux command line.  Full details are provided in the program
# documentation at the web site above.
#
# CUSTOMIZATION: Full details are provided in the program documentation
# at the web site above.
#
# The CasValue program invokes the bowtie sequence # mapping program.  For 
# information on bowtie, see
#
#  Langmead B, Trapnell C, Pop M, Salzberg SL.,
#     Ultrafast and memory-efficient alignment of short DNA sequences to
#     the human genome. Genome Biol 10:R25.
#
# AVAILABILITY AND CITATION
# 
# The CasFinder system of programs comprises CasFinder.pl, CasValue_v2.pl,
# the two perl packages CasFinderConfigFile.pm and CasFinderConfigCas9File.pm, 
# and the two configuration files CASFINDER_CONFIG.txt and 
# CASFINDER_CONFIG_CAS9.txt.
# 
# The CasFinder system of programs is being made available under the MIT 
# Licence (MIT).   
#
# Please cite the article: 
#
#   Aach J., Mali P., Church G.M..
#     CasFinder: Flexible algorithm for identifying specific Cas9 targets 
#     in genomes.  bioRxiv (doi)   
#
# when using, copying, or employing code from this system.
#
# COPYRIGHT NOTICE
# 
# The MIT License (MIT)
#
# Copyright (c) 2014 John Aach and the President and Fellows of Harvard University
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# PROGRAM CHANGE LOG:
# Created: 08/28/2013-05/12/2014 by John Aach
#
  use File::Spec;
  use Cwd 'abs_path';
  @INC=(".",@INC);
  use CasFinderConfigFile;  
  use CasFinderConfigCas9File;  

  ($vol,$dir,$PROGRAM)=File::Spec->splitpath(abs_path($0));
  $PGMPREF=$PROGRAM;
  $PGMPREF=~s/\.pl$//;
  
  # FLOWCHART NOTE:
  # The following call performs the logic in the box "Analyze parameters & retrieve config file info"
  AnalyzeParameters(@ARGV);
  
  # FLOWCHART NOTE:
  # The following call performs the logic in the box "Read input sequence; analyze FASTA keywords"
  ReadInputFile();
 
  # FLOWCHART NOTE:
  # The following call performs the logic in the box "Invoke bowtie to identify genome matches"
  PerformBowtieSeedSearch();

  # FLOWCHART NOTE:
  # The following call performs the logic in the box "Extract and analyze bowtie genome matches"
  ExtractAndAnalyzeGenomeHits();

  $outfile="$OUTPREF.data.txt";

  Msg("\nWriting output file $outfile.\n");
  
  # FLOWCHART NOTE:
  # The following code performs the logic in the box "Report as accepted all input Cas9 targets with <= 'accept' scores <= -x"
  $rejected_target_nonacgt=$accepted_target=0;
  open(O,">$outfile");
  print O "input_targetseq_index\tinput_targetseq_name\ttarget_threshold_rejection_score\tall_matches_analyzed\tmatches_with_nonstandard_bases\n";   
  for ($i=0;$i<@TARGETNAME;++$i){
     $all_matches_analyzed=1;
     if($BOWTIE_K>0 && $TOT_BOWTIE_RETURNS[$i]>=$BOWTIE_K){
        $all_matches_analyzed=0;
     }
     $score_array=$HIT_SCORE_ARRAY[$i];
     $score_array_size=@$score_array;
     if($score_array_size>$TARGETACCEPTNUM[$i]+1){
        die "Size error in score_array for target $TARGETNAME[$i] (ix=$i) (accept-num=$TARGETACCEPTNUM[$i]): array=(".(join",",@$score_array).").\n";
     }
     $accept_target=0;
     if($TARGETACCEPTNUM[$i]==0){
       if($score_array_size==0){
          $trs="Inf";
          $accept_target=1;
       }
       elsif($$score_array[0]>$TOTAL_SCORE_IGNORE_THRESHOLD){
          $trs=$$score_array[0];
          $accept_target=1;
       }
     }
     else{
       if($score_array_size<=$TARGETACCEPTNUM[$i]){
          $trs="Inf";
          $accept_target=1;
       }
       elsif($score_array_size==$TARGETACCEPTNUM[$i]+1){
          $trs=$$score_array[$score_array_size-1];
          if($trs>$TOTAL_SCORE_IGNORE_THRESHOLD){
            $accept_target=1;
          }
       }
     }
     if($DEBUG==1){
        print DEBUG "target_acceptance_data: tgtix=$i, tgtname=$TARGETNAME[$i], acceptnum=$TARGETACCEPTNUM[$i], score_array=(".(join",",@$score_array)."), accepted_ind=$accept_target\n";
     }
     if($accept_target==1){
        print O "$i\t$TARGETNAME[$i]\t$trs\t$all_matches_analyzed\t$TOT_HITS_NONSTANDARD_BASE[$i]\n";
        $accepted_target++;
     }
  }
  Msg("$accepted_target input target sequences reported as passing defined specificity conditions");

  Msg("\nProgram complete.\n");
  close I;
  close O;
  if($OUTFILE_SITE_DETAIL==1){
    close OUTDETAIL;
  }
  if($DEBUG==1){
    close DEBUG;
  }
    
# FLOWCHART NOTE:
# The following subroutine performs the logic in the box "Read input sequence; analyze FASTA keywords"
sub ReadInputFile{
  my $inseq=0;
  my $inreject=0;
  my $validseq;
  my $i;
  my $rest;
  my $seqno=-1;
  Msg("\nReading input target sequence file.\n");

  @TARGETSEQ=@TARGETNAME=@TARGETACCEPTNUM=();
  open(IF,$INFILE);
  while(<IF>){
    s/\s+$//;
    if(/^>(\S+)(.*)/){
      ++$seqno;
      $TARGETNAME[$seqno]=$1;
      $rest=$2;
      $TARGETSEQ[$seqno]="";
      $rest=$2;
      if($FORCE_ACCEPT eq ""){
        if($rest=~/accept=(\d+)/){
           $TARGETACCEPTNUM[$seqno]=$1;
        }
        else {
           $TARGETACCEPTNUM[$seqno]=1;
        }
      }
      else {
         $TARGETACCEPTNUM[$seqno]=$FORCE_ACCEPT;
      }
      ++$inseq;
    }
    else{
       $TARGETSEQ[$seqno].=uc($_);
    }
  }

  close IF;
  
  # validate sequences
  for($i=$#TARGETSEQ;$i>=0;$i--){
     if($TARGETSEQ[$i]=~/[^ACGT]/||length($TARGETSEQ[$i])!=$CURR_CAS9_TARGETLEN){
        ++$inreject;
        splice(@TARGETSEQ,$i,1);
        splice(@TARGETNAME,$i,1);
        splice(@TARGETACCEPTNUM,$i,1);
     }
     else {
       $validseq++;
     }
  }

  Msg("$inseq target sequences read from -i file.");
  MsgContinue("$inreject sequences rejected due to non-standard base codes or target length errors.");
  MsgContinue("$validseq sequences will be processed.\n");
  if($validseq==0){
     Msg("No input target sequences: Program terminating.\n");
     exit;
  }
  
}

# FLOWCHART NOTE:
# The following subroutine performs the logic in the box "Invoke bowtie to identify genome matches"
sub PerformBowtieSeedSearch{
  my $idnum_lnth;
  my $fmtstring;
  my $seedlen;
  my $i;
  my $seedname;
  my $bowtie_k_term;
  my $bowtie_command;
  my @bowtie_stderr_info=();

  $TEMP_BOWTIE_INFILE="$TEMPPREF.bowtie.in";
  $TEMP_BOWTIE_OUTFILE="$TEMPPREF.bowtie.out";
  $TEMP_BOWTIE_STDERR="$TEMPPREF.bowtie.stderr";
  Msg("Creating bowtie target seed file.");
  open(B,">$TEMP_BOWTIE_INFILE");
  $idnum_lnth=log((scalar @TARGETSEQ))/log(10);
  $idnum_lnth=($idnum_lnth==int($idnum_lnth)?$idnum_lnth:int($idnum_lnth)+1);
  $fmtstring="R\%0${idnum_lnth}d";
  $seedlen=$SEED_END-$SEED_START+1;
  for($i=0;$i<@TARGETSEQ;++$i){
    $seedname=sprintf($fmtstring,$i);
    print B ">$seedname\n".substr($TARGETSEQ[$i],$SEED_START-1,$seedlen)."\n";
  }
  
  # FLOWCHART NOTE:
  # The following code performs the logic in the box "Call bowtie with parameters based on -sm and -k options"  
  # This logic formats the bowtie command
  $bowtie_k_term="-a";
  if($BOWTIE_K>0){
    $bowtie_k_term="-k $BOWTIE_K";
  }
  $bowtie_command="$BOWTIE_EXECUTABLE $CURR_GENOME_BOWTIE_INDEX -l $seedlen -v $SEED_MISMATCH $bowtie_k_term -f $TEMP_BOWTIE_INFILE $TEMP_BOWTIE_OUTFILE 2> $TEMP_BOWTIE_STDERR"; 

  if($DEBUG==1){
    print DEBUG "bowtie command:\n$bowtie_command\n";
  }

  # This is the call to bowtie
  Msg("\nInitiating bowtie search");
  system($bowtie_command);
  Msg("\nBowtie search completed");
  
  open(S,$TEMP_BOWTIE_STDERR);
  while(<S>) { 
     s/\s+$//;
     push @bowtie_stderr_info, $_; 
  }
  close S;  
  if(@bowtie_stderr_info>0){
     MsgContinue("Bowtie output messages:",@bowtie_stderr_info);
  }

  # FLOWCHART NOTE:
  # The following code performs the logic in the box "Delete bowtie input file"  
  if($DEBUG!=1){  # delete files no longer needed (if not in debug mode)
    unlink $TEMP_BOWTIE_INFILE, $TEMP_BOWTIE_STDERR;
  }

}

# FLOWCHART NOTE:
# The following subroutine performs the logic in the box "Extract and analyze bowtie genome matches"
sub ExtractAndAnalyzeGenomeHits{
  # Note that genome hits are in $TEMP_BOWTIE_OUTFILE  
  my $i;
  my $seqname;
  my $seqix;    
  my $chr;
  my $leftend;
  my $start;
  my $end;
  my $name;
  my $strand;
  my $totlen=$CURR_CAS9_TARGETLEN+$CURR_CAS9_PAMLNTH; 
  my %chrom_extract_info=();  # hash by chromosome of bed file extract information 
  my $arr;  
  my $extract_data; 
  my $chrseq="";    
  my $chrhdr;       
  my $extract_start;  
  my $extract_end;    
  my $extract_strand; 
  my $hit_name;
  my $hit_location;
  my $hit_seq;
  my $hit_target;
  my $hit_pam;
  my $hit_accept_pam;
  my $hit_alternate_pam;
  my $hit_target_ix;
  my $hit_length_errs;
  my $hit_total_mismatches;
  my $hit_seed_mismatches;
  my $hit_nonseed_mismatches;
  my $hit_nonacgt;
  my $hit_score;
  my @hit_data;
  my $DEBUG_HITEXTRACT_LIMIT=1000;
  my $DEBUG_HITEXTRACT_IN_IX=0;
  my $DEBUG_HITEXTRACT_OUT_IX=0;
  my @debug_score_array_before;
  my $debug_score_array_after;
  my $DEBUG_SCORE_ARRAY_LIMIT=5000;
  my $DEBUG_SCORE_ARRAY_IX=0;
  my @bo=(); 

  # initialize counters that will be used to judge acceptability of evaluated targets
  for($i=0;$i<@TARGETSEQ;++$i){
     $TOT_BOWTIE_RETURNS[$i]=0;
     $TOT_HITS_NONSTANDARD_BASE[$i]=0;
     $HIT_SCORE_ARRAY[$i]=Array();
  }

  # FLOWCHART NOTE:
  # The following code performs the logic in the box "Collect genome bowtie match locations by chromosome"  
  Msg("Reading target seed match information from bowtie output\n");  
  open(B,$TEMP_BOWTIE_OUTFILE);
  while(<B>){
    s/\s+$//;
    @bo=split(/\t/); 
    $seqname=$bo[0]; 
    $seqix=int(substr($seqname,1));  
    $TOT_BOWTIE_RETURNS[$seqix]++;   
    $strand=$bo[1];  
    $chr=$bo[2];     
    $leftend=$bo[3]; # NOTE: left end is a position with start index 0; 
    if($chrom_extract_info{$chr} eq"" && $$CURR_GENOME_CHROMOSOME_FILES{$chr}ne""){ # skip miscellaneous chromosomes that might be in bowtie index (e.g., mitochondrial chromomomes)
       $chrom_extract_info{$chr}=Array();
    }
    $arr=$chrom_extract_info{$chr};
    if($strand eq "+"){
      $start=$leftend-($SEED_START-1);  # start is a position with index 0
      if($start<0) {next;} # no point in retrieving this
    }
    else{
      $end=$leftend+$SEED_END;
      $start=$end-$totlen;
      if($start<0) {next;} # no point in retrieving this
    }

    $extract_data=(join"\t",($start,$strand,$seqname)); 
    push @$arr,$extract_data; 

    if($DEBUG==1){
      if($DEBUG_HITEXTRACT_IN_IX<$DEBUG_HITEXTRACT_LIMIT){
        if($DEBUG_HITEXTRACT_IN_IX==0){
            print DEBUG "hit extract in data: first $DEBUG_HITEXTRACT_LIMIT entries:\n";
        }
        ++$DEBUG_HITEXTRACT_IN_IX;
        print DEBUG "hit_extract_in_$DEBUG_HITEXTRACT_IN_IX\t$chr\t$extract_data\n";
      }
    }
  }
  close B;

  # FLOWCHART NOTE:
  # The following code performs the logic in the box "Delete bowtie output file"  
  if($DEBUG!=1){  # delete files no longer needed (if not in debug mode)
    unlink $TEMP_BOWTIE_OUTFILE;
  }

  # FLOWCHART NOTE:
  # The following code performs the logic in the box "For each chromosome, extract full Cas9 footprint around each bowtie match, reject non-PAM footprints, and score the rest"
  Msg("Extracting sequences around target seed matches from chromosome files.\n");  
  foreach $chr(@$CURR_GENOME_CHROMOSOME_ORDER){
    if($chrom_extract_info{$chr} eq "") {next;}
    $arr=$chrom_extract_info{$chr};
  
    open(C,$$CURR_GENOME_CHROMOSOME_FILES{$chr});
    $chrhdr=<C>;
    $chrseq="";
    while(<C>){
       s/\s+$//;
       if(/^>/){
          Msg("\nUnexpected FASTA header in chromosome file $$CURR_GENOME_CHROMOSOME_FILES{$chr}:");
          MsgContinue("$_");
          MsgContinue("Program terminating.");
          exit;
       }
       $chrseq.=$_;
    }
    close C; 

    foreach $extract_data(@$arr){
      ($extract_start,$extract_strand,$hit_name)=split/\t/,$extract_data;  
      $hit_seq=substr($chrseq,$extract_start,$totlen);          
      $hit_seq=uc($hit_seq);
      if($extract_strand eq"-"){
         $hit_seq=RC($hit_seq);
      }
      $extract_end=$extract_start+$totlen;   
      $hit_location="$chr:".($extract_start+1)."-$extract_end:$extract_strand";  
      if($DEBUG==1){
         if($DEBUG_HITEXTRACT_OUT_IX<$DEBUG_HITEXTRACT_LIMIT){
           if($DEBUG_HITEXTRACT_OUT_IX==0){
             print DEBUG "hit extract out data: first $DEBUG_HITEXTRACT_LIMIT records:\n";
           }
           print DEBUG "hit_extract_out_".($DEBUG_HITEXTRACT_OUT_IX+1)."\t$chr\t$hit_name\t$hit_location\t$hit_seq";
         }
         ++$DEBUG_HITEXTRACT_OUT_IX;
      }
      
      if(length($hit_seq) != $totlen){$hit_length_errs++; next;} # could happen for hits at end of chromosome
      $hit_target_ix=int(substr($hit_name,1));
      $hit_target=substr($hit_seq,0,$CURR_CAS9_TARGETLEN);
      $hit_pam=substr($hit_seq,$CURR_CAS9_TARGETLEN);
      $hit_accept_pam=0;
      $hit_alternate_pam=0;
      $hit_total_mismatches=0;
      $hit_seed_mismatches=0;
      $hit_nonseed_mismatches=0;
      $hit_nonacgt=0;
      $hit_score=0;
      if($hit_pam=~/^$CURR_CAS9_PAM_REGEXP$/){  
        $hit_accept_pam=1;          
      }
      elsif($hit_pam=~/^$CURR_CAS9_PAM2_REGEXP$/){   
        $hit_accept_pam=1;          
        $hit_alternate_pam=1;       
      }                             
      else{
        if($DEBUG==1 && $DEBUG_HITEXTRACT_OUT_IX<=$DEBUG_HITEXTRACT_LIMIT){
          print DEBUG "\t$hit_name\tNON-PAM\n";
        }
        next;
      }
      if($hit_seq=~/[^ACGT]/){
          $hit_nonacgt=1;
      }
      ($hit_total_mismatches,$hit_seed_mismatches,$hit_nonseed_mismatches)=Mismatches($TARGETSEQ[$hit_target_ix],$hit_target);
      $hit_score=$hit_seed_mismatches*$SEED_MISMATCH_SCORE+$hit_nonseed_mismatches*$NONSEED_MISMATCH_SCORE+$hit_alternate_pam*$ALTERNATE_PAM_SCORE;
      if($OUTFILE_SITE_DETAIL==1 || ($DEBUG==1 && $DEBUG_HITEXTRACT_OUT_IX<=$DEBUG_HITEXTRACT_LIMIT)){
        @hit_data=($hit_score,$hit_total_mismatches,$hit_alternate_pam,$hit_seed_mismatches,$hit_nonseed_mismatches,$hit_nonacgt);
      }
      if($DEBUG==1 && $DEBUG_HITEXTRACT_OUT_IX<=$DEBUG_HITEXTRACT_LIMIT){
          print DEBUG "\t(score,totmm,altpam,seedmm,nonseedmm,nonacgt)=(".(join",",@hit_data).")\n";
      }
      if($hit_score<=$TOTAL_SCORE_IGNORE_THRESHOLD){
        if($OUTFILE_SITE_DETAIL==1){
            print OUTDETAIL (join"\t",($hit_target_ix,$TARGETNAME[$hit_target_ix],$hit_location,@hit_data,$hit_seq));
            print OUTDETAIL "\n";
        }
      }
      if($hit_nonacgt==1){
        $TOT_HITS_NONSTANDARD_BASE[$hit_target_ix]++;
      }
      if($DEBUG==1 && $DEBUG_SCORE_ARRAY_IX<=$DEBUG_SCORE_ARRAY_LIMIT){
         @debug_score_array_before=@{$HIT_SCORE_ARRAY[$hit_target_ix]};
      }
      InsertScoreIntoArray($hit_score,$HIT_SCORE_ARRAY[$hit_target_ix],$TARGETACCEPTNUM[$hit_target_ix]);
      if($DEBUG==1 && $DEBUG_SCORE_ARRAY_IX<=$DEBUG_SCORE_ARRAY_LIMIT){
         $debug_score_array_after=$HIT_SCORE_ARRAY[$hit_target_ix];
         print DEBUG "score_array_update_$DEBUG_SCORE_ARRAY_IX\t$hit_name\tbefore=(".(join",",@debug_score_array_before).
                   ")\tnewscore=$hit_score\tafter=(".(join",",@$debug_score_array_after).").\n";
         ++$DEBUG_SCORE_ARRAY_IX;
      }
      
      delete $chrom_extract_info{$chr};   # Get rid of unused storage (statement moved)
    }

  }  

}

sub InsertScoreIntoArray{
  my $hit_score=$_[0];
  my $score_array=$_[1];
  my $accept_num=$_[2];
  my $score_array_size=(scalar @$score_array);
  my $curr_array_maxindex=$score_array_size-1;
  my $i;
  my $place_score_ix;
  if($curr_array_maxindex<0){
     $$score_array[0]=$hit_score;
     return;
  }
  if($score_array_size>=$accept_num+1 && $hit_score>=$$score_array[$curr_array_maxindex]){
    return;
  }
  $i=$curr_array_maxindex;
  $place_score_ix=$i+1;
  while($i>=0&&$$score_array[$i]>$hit_score){
     unless($i==$curr_array_maxindex&&$curr_array_maxindex>=$accept_num){
       $$score_array[$i+1]=$$score_array[$i];
     }
     $place_score_ix=$i;
     --$i;
   }
   $$score_array[$place_score_ix]=$hit_score;
}

sub Mismatches{
  my @seq1=split//,$_[0];
  my @seq2=split//,$_[1];
  my $i;
  my $tmm=0;
  my $smm=0;
  my $nmm=0;
  if((scalar @seq1) != (scalar @seq2)){
     die "Error in Mismatches routine, length mismatch. Seq1=$_[0], Seq2=$_[1]\n";
  }
  for($i=0;$i<@seq1;++$i){
     if($seq1[$i]ne$seq2[$i]){
        ++$tmm;
        if($i>=$SEED_START-1&&$i<$SEED_END){
           ++$smm;
        }
        else{
           ++$nmm;
        }
     }
  }
  return ($tmm,$smm,$nmm);
}

sub Hash{
  my %h=();
  return \%h;
}

sub Array{
  my @a=();
  return \@a;
}

sub TimeStamp {
  my $large = $_[0];
  my $ts = "";
  my $yr = (localtime(time()))[5]+1900;
  my $mo = (localtime(time()))[4];
  my $da = (localtime(time()))[3];
  my $hr = (localtime(time()))[2];
  my $mi = (localtime(time()))[1];
  my $se = (localtime(time()))[0];

  if ($large eq "LARGE") {   
    $mo = (Jan,Feb,Mar,Apr,May,Jun,Jul,Aug,Sep,Oct,Nov,Dev)[$mo];
    $ts = "$mo $da, $yr at ";
  }
  if ($hr<10){$hr = "0$hr";} 
  if ($mi<10){$mi = "0$mi";} 
  if ($se<10){$se = "0$se";} 
  return "$ts$hr:$mi:$se";
}

sub Msg{
  my $msg=$_[0];
  if($msg=~/^(\n+)/){
     print I $1;
     $msg=~s/^\n+//;
  }
  print I TimeStamp().": $msg\n";
}
sub MsgContinue{
  my $msg;
  foreach $msg(@_){
    print I " "x(length(TimeStamp())+2);
    print I "$msg\n";
  }
}

sub RC{
  my @s=split//,$_[0]; 
  my $rc=join("",(reverse @s));
  $rc=~tr/ACGT/TGCA/;
  return $rc;
}

# FLOWCHART NOTE:
# The following subroutine performs the logic in the box "Analyze parameters & retrieve config file info"
sub AnalyzeParameters{
  use IO::Handle;
  use File::Glob;
  my $i;
  my $a;
  my $h;
  my $chrfile;
  my @bowtie_index_files=();

  $INFILE="";
  $OUTPREF="";
  $CONFIG_FILE="";
  $CONFIG_CAS9_FILE="";
  $CONFIG_FILE_DFLT="CASFINDER_CONFIG.txt";  
  $CONFIG_CAS9_FILE_DFLT="CASFINDER_CAS9_CONFIG.txt";  
  $CAS9="";
  $CAS9_DFLT="SP";
  $GENOME="";
  $GENOME_DFLT="hg19";
  $SEED_START="";
  $SEED_END="";
  $SEED_MISMATCH="";
  $SEED_MISMATCH_DFLT=1; # -sm max number of seed mismatches considered
  $TOTAL_SCORE_IGNORE_THRESHOLD=""; 
  $TOTAL_SCORE_IGNORE_THRESHOLD_DFLT=3; # -x score threshold above which a site is ignored as not being a significant 
  $ALTERNATE_PAM_SCORE="";
  $ALTERNATE_PAM_SCORE_DFLT=1; # -asc score for hit with alternate PAM 
  $SEED_MISMATCH_SCORE="";
  $SEED_MISMATCH_SCORE_DFLT=1; # -ssc score for each hit seed mismatch
  $NONSEED_MISMATCH_SCORE="";
  $NONSEED_MISMATCH_SCORE_DFLT=1; # -nsc score for each non-seed mismatch
  $BOWTIE_K="";
  $BOWTIE_K_DFLT=0;  # if 0, run bowtie with -a, otherwise with -k 
  $OUTFILE_SITE_DETAIL=0; # -D parameter.  If set to 1, causes file to be written with locations and score details for each site hit within score threshold
  $DEBUG=0;
  $FORCE_ACCEPT="";

  for($i=0;$i<@_;++$i){
     if($_[$i] eq"-i"){
        if($i==$#_){
           die "No value specified for parameter $_[$i]\n";
        }
        $INFILE=$_[$i+1];
        ++$i;
     }
     elsif($_[$i] eq"-o"){
        if($i==$#_){
           die "No value specified for parameter $_[$i]\n";
        }
        $OUTPREF=$_[$i+1];
        ++$i;
     }
     elsif($_[$i] eq"-CONFIG"){
        if($i==$#_){
           die "No value specified for parameter $_[$i]\n";
        }
        $CONFIG_FILE=$_[$i+1];
        ++$i;
     }
     elsif($_[$i] eq"-CONFIG_CAS9"){
        if($i==$#_){
           die "No value specified for parameter $_[$i]\n";
        }
        $CONFIG_CAS9_FILE=$_[$i+1];
        ++$i;
     }
     elsif($_[$i] eq"-k"){
        if($i==$#_){
           die "No value specified for parameter $_[$i]\n";
        }
        unless($_[$i+1]=~/^\d+$/i) {  
           die "Invalid value specified for $_[$i] parameter.\n";
        }
        $BOWTIE_K=$_[$i+1];
        ++$i;
     }
     elsif($_[$i] eq"-A"){
        if($i==$#_){
           die "No value specified for parameter $_[$i]\n";
        }
        unless($_[$i+1]=~/^\d+$/i) {
           die "Invalid value specified for $_[$i] parameter.\n";
        }
        $FORCE_ACCEPT=$_[$i+1];
        ++$i;
     }
     elsif($_[$i] eq"-c"){
        if($i==$#_){
           die "No value specified for parameter $_[$i]\n";
        }
        $CAS9=$_[$i+1];
        ++$i;
     }
     elsif($_[$i] eq"-g"){
        if($i==$#_){
           die "No value specified for parameter $_[$i]\n";
        }
        $GENOME=$_[$i+1];
        ++$i;
     }
     elsif($_[$i] eq"-ss"){
        if($i==$#_){
           die "No value specified for parameter $_[$i]\n";
        }
        $SEED_START=$_[$i+1];
        unless($SEED_START=~/^\d+/ && $SEED_START>0){
           die "Invalid value specified for $_[$i] parameter.\n";
        }
        ++$i;
     }
     elsif($_[$i] eq"-se"){
        if($i==$#_){
           die "No value specified for parameter $_[$i]\n";
        }
        $SEED_END=$_[$i+1];
        unless($SEED_END=~/^\d+/ && $SEED_END>0){
           die "Invalid value specified for $_[$i] parameter.\n";
        }
        ++$i;
     }
     elsif($_[$i] eq"-sm"){
        if($i==$#_){
           die "No value specified for parameter $_[$i]\n";
        }
        $SEED_MISMATCH=$_[$i+1];
        unless($SEED_MISMATCH=~/^\d+/){
           die "Invalid value specified for $_[$i] parameter.\n";
        }
        ++$i;
     }
     elsif($_[$i] eq"-x"){
        if($i==$#_){
           die "No value specified for parameter $_[$i]\n";
        }
        $TOTAL_SCORE_IGNORE_THRESHOLD=$_[$i+1];
        unless($TOTAL_SCORE_IGNORE_THRESHOLD=~/^[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?+/ && $TOTAL_SCORE_IGNORE_THRESHOLD>0){
           die "Invalid value specified for $_[$i] parameter.\n";
        }
        ++$i;
     }
     elsif($_[$i] eq"-ssc"){
        if($i==$#_){
           die "No value specified for parameter $_[$i]\n";
        }
        $SEED_MISMATCH_SCORE=$_[$i+1];
        unless($SEED_MISMATCH_SCORE=~/^[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?$/ && $SEED_MISMATCH_SCORE>0){
           die "Invalid value specified for $_[$i] parameter.\n";
        }
        ++$i;
     }
     elsif($_[$i] eq"-nsc"){
        if($i==$#_){
           die "No value specified for parameter $_[$i]\n";
        }
        $NONSEED_MISMATCH_SCORE=$_[$i+1];
        unless($NONSEED_MISMATCH_SCORE=~/^[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?$/ && $NONSEED_MISMATCH_SCORE>0){
           die "Invalid value specified for $_[$i] parameter.\n";
        }
        ++$i;
     }
     elsif($_[$i] eq"-asc"){
        if($i==$#_){
           die "No value specified for parameter $_[$i]\n";
        }
        $ALTERNATE_PAM_SCORE=$_[$i+1];
        unless($ALTERNATE_PAM_SCORE=~/^[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?+/){
           die "Invalid value specified for $_[$i] parameter.\n";
        }
        ++$i;
     }
     elsif($_[$i] eq "-D"){
       $OUTFILE_SITE_DETAIL=1;
     }
     elsif($_[$i] eq "-DEBUG"){
       $DEBUG=1;
     }
     elsif($_[$i]=~/^-\?/ || $_[$i]=~/^-h(e|el|elp)?$/i){
        print "$PROGRAM parameters:\n";
        print " -CONFIG: CasValue configuration file (default=$CONFIG_FILE_DFLT)\n";
        print " -CONFIG_CAS9: CasValue Cas9 configuration file (default=$CONFIG_CAS9_FILE_DFLT)\n";
        print " -i:  input FASTA file of target sequences (see NOTE (1)) (required)\n";
        print " -o:  output file prefix (required)\n";
        print " -c:  Cas9 specifier (default=$CAS9_DFLT)\n";
        print " -g:  target genome (default=$GENOME_DFLT)\n";
        print " -k:  bowtie -k parameter (0 use bowtie -a, >0 use bowtie -k)(default=$BOWTIE_K_DFLT)\n";
        print " -sm: max number of mismatches to seed considered (default=$SEED_MISMATCH_DFLT)\n";
        print " -ss: override of Cas9 seed start position (default in -CONFIG_CAS9 file)\n";
        print " -se: override of Cas9 seed end position (default in -CONFIG_CAS9 file)\n";
        print " -x:  score threshold above which sites are ignored (default=$TOTAL_SCORE_IGNORE_THRESHOLD_DFLT)\n";
        print " -asc: score for alternate PAM (default=$ALTERNATE_PAM_SCORE_DFLT)\n";
        print " -ssc: score for seed mismatch (default=$SEED_MISMATCH_SCORE_DFLT)\n";
        print " -nsc: score for non-seed mismatch (default=$NONSEED_MISMATCH_SCORE_DFLT)\n";
        print " -A:  if used, forces all input sequence accept-numbers to specified value\n";
        print "      (see NOTE (1))\n";
        print " -D:  print output site detail file (see NOTE (2)) (optional)\n";
        print "NOTES:\n";
        print " (1) By default, $PROGRAM will report only targets which contain at most 1\n";
        print "     matching site in the genome with score <= -x.  This may be overridden\n";
        print "     by including a keyword matching accept=\\d+ in the FASTA header of\n";
        print "     a target sequence in the -i file.  For instance, accept=0 will cause\n";
        print "     a target to be reported only if there are no matches in the genome\n";
        print "     with score <= -x.  This would be appropriate for a target in an\n";
        print "     introduced exogenous sequence.  Use of the optional -A parameter \n";
        print "     overrides all individual input sequence-specified accept numbers.\n"; 
        print " (2) -D parameter causes $PROGRAM to print a file detailing all genome\n";
        print "     matches with scores <= -x for all targets in the input file. This\n";
        print "     will be voluminous if large numbers of targets are processed.  The\n";
        print "     output detail file will not be sorted by input target.\n";
        exit;
     }
     else{
        die "Unrecognized parameter $_[$i]\n";
     }
  }
  if($CONFIG_FILE eq""){
    $CONFIG_FILE=$CONFIG_FILE_DFLT;
  }
  if($CONFIG_CAS9_FILE eq""){
    $CONFIG_CAS9_FILE=$CONFIG_CAS9_FILE_DFLT;
  }
  if($GENOME eq""){
    $GENOME=$GENOME_DFLT;
  }
  if($CAS9 eq""){
    $CAS9=$CAS9_DFLT;
  }
  if($BOWTIE_K eq""){
    $BOWTIE_K=$BOWTIE_K_DFLT;
  }
  if($SEED_MISMATCH eq""){
    $SEED_MISMATCH=$SEED_MISMATCH_DFLT;
  }
  if($TOTAL_SCORE_IGNORE_THRESHOLD eq""){
    $TOTAL_SCORE_IGNORE_THRESHOLD=$TOTAL_SCORE_IGNORE_THRESHOLD_DFLT;
  }
  if($ALTERNATE_PAM_SCORE eq""){
    $ALTERNATE_PAM_SCORE=$ALTERNATE_PAM_SCORE_DFLT;
  }
  if($SEED_MISMATCH_SCORE eq""){
    $SEED_MISMATCH_SCORE=$SEED_MISMATCH_SCORE_DFLT;
  }
  if($NONSEED_MISMATCH_SCORE eq""){
    $NONSEED_MISMATCH_SCORE=$NONSEED_MISMATCH_SCORE_DFLT;
  }
  if($INFILE eq ""){
     die "No input target file specified (-i)\n";
  }
  if($OUTPREF eq ""){
     die "No output file prefix specified (-o)\n";
  }
  if($SEED_START ne "" && $SEED_END ne "" && $SEED_START >= $SEED_END){
     die "Inconsistent values of -ss and -se parameters\n";
  }
  if($SEED_MISMATCH>3){
     die "Unsupported -sm value: bowtie does not support searches wtih >3 mismatches.\n";
  }
  if(! -e $CONFIG_FILE){
     die "Specified general configuration file -CONFIG does not exist.\n";
  }
  if(! -e $CONFIG_CAS9_FILE){
     die "Specified Cas9 configuration file -CONFIG does not exist.\n";
  }
  if(! -e $INFILE){
     die "Specified input file -i does not exist.\n";
  }
  open(I,">$OUTPREF.info.txt");
  autoflush I 1;
  
  print I "Parameters for execution of $PROGRAM on ".TimeStamp("LARGE").":\n";
  print I "-CONFIG: $CONFIG_FILE (general configuration file)\n";  
  print I "-CONFIG_CAS9: $CONFIG_CAS9_FILE (Cas9 configuration file)\n";
  print I "-i:   $INFILE (input target sequence file)\n";
  print I "-o:   $OUTPREF (output file prefix)\n";
  print I "-c:   $CAS9 (Cas9 version)\n";
  print I "-g:   $GENOME (target genome)\n";
  print I "-k:   $BOWTIE_K (bowtie -k parameter)\n";
  print I "-sm:  $SEED_MISMATCH (max number of seed mismatches considered)\n";
  if($SEED_START ne ""){
    print I "-ss:  $SEED_START (target seed region start position)\n";
  }
  else {
    print I "-ss:  target seed start not specified: will default to -CONFIG_CAS9 default.\n";
  }
  if($SEED_END ne ""){
    print I "-se:  $SEED_END (target seed region end position)\n";
  }
  else {
    print I "-se:  target seed end not specified: will default to -CONFIG_CAS9 default.\n";
  }
  print I "-x:   $TOTAL_SCORE_IGNORE_THRESHOLD (score threshold above which sites are ignored)\n";
  print I "-asc: $ALTERNATE_PAM_SCORE (score for alternate PAM)\n";
  print I "-ssc: $SEED_MISMATCH_SCORE (score for seed mismatch)\n";
  print I "-nsc: $NONSEED_MISMATCH_SCORE (score for non-seed mismatch)\n";
  if($OUTFILE_SITE_DETAIL==1){  
    print I "-D:   output site_detail file will be generated.\n";
  }
  if($FORCE_ACCEPT ne ""){
    print I "-A:   $FORCE_ACCEPT (accept number forced for each input sequence)\n";
  }   
  print I "\n";

  if($DEBUG==1){
    my $dbgfile="$PROGRAM.DEBUG.out";
    open(DEBUG,">$dbgfile");
    Msg("\nProgram is in DEBUG mode.\n");
  }

  # Now read and integrate wtih configuration files
  CasFinderConfigFile->Read($CONFIG_FILE); 
  CasFinderConfigCas9File->Read($CONFIG_CAS9_FILE); 
 
  $CURR_GENOME_BOWTIE_INDEX=CasFinderConfigFile->Get("GENOME_BOWTIE_INDEX",$GENOME);
  $CURR_GENOME_CHROMOSOME_ORDER=CasFinderConfigFile->Get("GENOME_CHROMOSOME_ORDER",$GENOME);
  $CURR_GENOME_CHROMOSOME_FILES=CasFinderConfigFile->Get("GENOME_CHROMOSOME_FILES",$GENOME);
  $BOWTIE_EXECUTABLE=CasFinderConfigFile->Get("BOWTIE_EXECUTABLE");

  $TEMPPREF=CasFinderConfigFile->Get("TEMPPREF");
  if($DEBUG==1){  # if in debug mode, reset temporary prefix
    $TEMPPREF="testtemp";
  }

  if($CURR_GENOME_BOWTIE_INDEX eq"" || $CURR_GENOME_CHROMOSOME_ORDER eq"" || $CURR_GENOME_CHROMOSOME_FILES eq""){
    print I "\nSpecified genome $GENOME not specified in configuration file $CONFIG_FILE\n";
    print I "Program terminating.\n";
    exit;
  }
 
  # check for existence of bowtie and chromsome files
  @bowtie_index_files=<$CURR_GENOME_BOWTIE_INDEX*.ebwt>;  # File::Glob
  if(@bowtie_index_files==0){
    print I "\nSpecified bowtie index file for $GENOME in configuration file $CONFIG_FILE does not exist:\n$CURR_GENOME_BOWTIE_INDEX\n";
    print I "Program terminating.\n";
    exit;
  }
  foreach $chrfile(@$CURR_GENOME_CHROMOSOME_ORDER){
    if(! -e $$CURR_GENOME_CHROMOSOME_FILES{$chrfile}){ 
      print I "\nSpecified chromosome file for $chrfile for $GENOME in configuration file $CONFIG_FILE does not exist:\n$$CURR_GENOME_CHROMOSOME_FILES{$chrfile}\n";
      print I "Program terminating.\n";
      exit;
    }
  }
  Msg("System parameters read from configuration file $CONFIG_FILE:");
  MsgContinue("Bowtie executable = $BOWTIE_EXECUTABLE");
  MsgContinue("Temporary file prefix = $TEMPPREF\n");
  Msg("Bowtie index and chromosome files for genome $GENOME have been read from configuration file $CONFIG_FILE\n");

  $CURR_CAS9_TARGETLEN=CasFinderConfigCas9File->Get("CAS9_TARGETLEN",$CAS9);
  $CURR_CAS9_SEED_START=CasFinderConfigCas9File->Get("CAS9_SEED_START",$CAS9);
  $CURR_CAS9_SEED_END=CasFinderConfigCas9File->Get("CAS9_SEED_END",$CAS9);
  $CURR_CAS9_PAM=CasFinderConfigCas9File->Get("CAS9_PAM",$CAS9);
  $CURR_CAS9_PAMLNTH=CasFinderConfigCas9File->Get("CAS9_PAMLNTH",$CAS9);  
  $CURR_CAS9_PAM_REGEXP=CasFinderConfigCas9File->Get("CAS9_PAM_REGEXP",$CAS9);
  $CURR_CAS9_PAM2=CasFinderConfigCas9File->Get("CAS9_PAM2",$CAS9);
  $CURR_CAS9_PAM2_REGEXP=CasFinderConfigCas9File->Get("CAS9_PAM2_REGEXP",$CAS9);

  if($CURR_CAS9_TARGETLEN eq""||$CURR_CAS9_SEED_START eq""||$CURR_CAS9_SEED_END eq""||$CURR_CAS9_PAM eq""||$CURR_CAS9_PAM_REGEXP eq""){
    print I "Specified Cas9 $CAS9 not defined in configuration file $CONFIG_CAS9_FILE\n";
    print I "Program terminating.\n";
    exit;
  }
  
  Msg("\nCas9 $CAS9 characteristics read from configuration file $CONFIG_CAS9_FILE:");
  MsgContinue("target_length=$CURR_CAS9_TARGETLEN","seed_start=$CURR_CAS9_SEED_START","seed_end=$CURR_CAS9_SEED_END",
              "PAM=$CURR_CAS9_PAM ($CURR_CAS9_PAM_REGEXP)","PAM2=$CURR_CAS9_PAM2 ($CURR_CAS9_PAM2_REGEXP)",  
              "PAMLNTH=$CURR_CAS9_PAMLNTH\n");  

  if($SEED_START eq ""){ 
    $SEED_START=$CURR_CAS9_SEED_START;
    Msg("\nTarget seed start position (-ss) set to -CONFIG_CAS9 default $SEED_START");
  }
  else {
    Msg("\nTarget seed start position (-ss) has overriden the -CONFIG_CAS9 default $CURR_CAS9_SEED_START");
  }
  if($SEED_END eq ""){ 
    $SEED_END=$CURR_CAS9_SEED_END;
    Msg("\nTarget seed end position (-se) set to -CONFIG_CAS9 default $SEED_END");
  }
  else {
    Msg("\nTarget seed end position (-se) has overriden the -CONFIG_CAS9 default $CURR_CAS9_SEED_END");
  }
  if($SEED_START>$CURR_CAS9_TARGETLEN || $SEED_END>$CURR_CAS9_TARGETLEN || $SEED_START>=$SEED_END){
     print I "Error: inconsistent -ss or -se override of Cas9 seed start or end parameters.\n";
     print I "Program terminating.\n";
     exit;
  } 
  
  if($OUTFILE_SITE_DETAIL==1){ # open output detail file
    $outdetail_file="$OUTPREF.site_detail.txt";
    Msg("\nOpening site detail file $outdetail_file for output.\n");
    open(OUTDETAIL,">$outdetail_file");
    print OUTDETAIL "target_index\ttarget_name\tsite_location\tsite_score\ttotal_mismatches\talternate_pam?\tseed_mismatches\tnonseed_mismatches\tnon_acgt_basecode?\tsite_sequence\n";
  }
  
}
